# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


from django.contrib import admin

from models import DataTypeNameSpace,\
    InfoObjectFamily,\
    Revision,\
    FactTerm2Type,\
    FactDataType,\
    InfoObjectType,\
    NodeID,\
    InfoObjectNaming,\
    BlobStorage,\
    IdentifierNameSpace



#
# Inline Interfaces
# -----------------
#
# Django offers the possibility to enrich an admin
# interfaces with admin areas for related objects
# that are 'inlined' into the main interface.
# To achieve this, Inline-classes have to
# be defined.
#
# We use the following naming convention:
#
# XXXXXXX[_zzzzzz_][_]YYYYYYYInline
#
# means that object XXXXXX contains an inline for object YYYYYYYY.
# 'zzzzzz' may be used if the same YYYYYY object is inlined in several
# ways according to multiple relations between XXXXXX and YYYYYY -- see
# examples below).  Underbars may be used to separate names where
# camel-casing gets to confusing.
#
# In inline interface, use the properties 'verbose_name' and 'verbose_name_plural'
# to provide information about the way in which inlines are related to the
# main object.
#
#


class InfoObjectType_InfoObjectNaming_Inline(admin.TabularInline):
    model = InfoObjectNaming
    extra = 0
    fields=('format_string','position')
    sortable_field_name = 'position'


class DataTypeNameSpace_FactDataType_Inline(admin.TabularInline):
    model = FactDataType
    extra = 0
    fields=('name','description')
    readonly_fields = ('description',)






#
# Admin Interfaces
# ----------------
#
# Below we specify admin interfaces in which
# we tweak the behavior of the standard admin
# interface:
#
# - list_display: which fields to display in the list of objects
# - list_filter: which fields can be used for filtering the list of objects
# - inlines: which admin interfaces should be inlined?
#
# We also hook into the save-on-change/create mechanism
# to do additional changes where necessary.
#
#

class FactDataTypeAdmin(admin.ModelAdmin):
    list_display = ('name','kind','namespace')
    raw_id_fields = ('namespace',)
    autocomplete_lookup_fields = {
        'fk': ['namespace'],
        'm2m': [],
    }


class DataTypeNameSpaceAdmin(admin.ModelAdmin):
    list_display = ('uri','name',)
    #filter_horizontal = ('factdatatype_set',)
    #inlines = (DataTypeNameSpace_FactDataType_Inline,)

class InfoObjectTypeAdmin(admin.ModelAdmin):
    list_display = ('name','iobject_family','namespace')
    inlines = (InfoObjectType_InfoObjectNaming_Inline,)

class FactTerm2TypeAdmin(admin.ModelAdmin):
    list_display = ('fact_term','iobject_type')
    raw_id_fields = ('iobject_type','fact_data_types')

    autocomplete_lookup_fields = {
        'fk': ['iobject_type',],
        'm2m': ['fact_data_types',],
        }



class InfoObjectFamilyAdmin(admin.ModelAdmin):
    list_display = ('name','title','description',)

class BlobStorageAdmin(admin.ModelAdmin):
    list_display = ('sha256',)

#
# Registration
# ------------
#
# Below, we register admin interfaces.
#

# Enumerables; useful for managing enumerables


admin.site.register(FactDataType,FactDataTypeAdmin)
admin.site.register(DataTypeNameSpace,DataTypeNameSpaceAdmin)
admin.site.register(IdentifierNameSpace)
admin.site.register(InfoObjectType,InfoObjectTypeAdmin)
admin.site.register(InfoObjectFamily,InfoObjectFamilyAdmin)
admin.site.register(FactTerm2Type,FactTerm2TypeAdmin)



# Helper object: Admin interface useful for checking, debugging...

admin.site.register(NodeID)
admin.site.register(Revision)
admin.site.register(BlobStorage,BlobStorageAdmin)






# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#



import logging
import pprint
import re
import hashlib
import base64


from django.db import models
from django.db.models import Count, F
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic
from django.core import urlresolvers
from django.utils.safestring import mark_safe
from django.utils import timezone
from django.core.files.base import ContentFile

import dingos.read_settings

import dingos

from dingos import *

from dingos.core.datastructures import DingoObjDict,ExtendedSortedDict

logger = logging.getLogger(__name__)
pp = pprint.PrettyPrinter(indent=2)

dingos_class_map = {}




RE_SEARCH_PLACEHOLDERS = re.compile(r"\[(?P<bla>[^\]]+)\]")


class DingoModel(models.Model):
    """
    This base class for all dingos models is used to provide all model classes
    with a dictionary of class definitions. The idea is that instead of
    using::

         SomeClass.objects.get_or_create(...)/filter(...)/...

    to access an instance, we use::

         self._DCM['SomeClass'].objects.get_or_create(...)/filter(...)/...

    As long, as we use Dingo proper, this is pointless, but it may be
    helpful when defining abstract classes based on Dingo classes:
    what we want when using abstract classes to always create and work
    with the abstract classes rather than the base Dingo classes.

    This approach does not quite work, because in many cases, instances
    are accessed via the class methods which, by default, always lead
    back to the base Dingo classes. For a while it seemed that this
    could be overcome with Django model managers, but at least until
    Django 1.5, there seems no way to force the usage of a user-defined
    model manager in all cases.

    """

    def __init__(self, *args, **kwargs):
        self._DCM = kwargs.get('dingos_class_map', dingos_class_map)
        try:
            del (kwargs['dingos_class_map'])
        except:
            pass
        models.Model.__init__(self, *args, **kwargs)


    class Meta:
        abstract = True


class FactValue(DingoModel):
    """
    All kind of information eventually boils down to a collection of facts. In DINGO,
    we keep all values of such facts in a single table/model, no matter whether
    the value is something completely unique/specific such as a hash value
    or belongs to a predefined data-type of value vocabulary.
    ("Yes" and "No"; "male", "female", "other", ...).

    If the fact value is part of a pre-defined vocabulary, we
    can store also a additional information about the value (title and description).
    the standard use case for this is to provide a human-readable value and
    help text for automatically generate forms in which the user can
    provide input (drop-down, radio-buttons, etc.)

    Finally: each fact value belongs to a fact data type. The consequence
    of this is that the same value will appear several times in the
    table if it belongs to several data types (e.g., "Yes" may be
    part of a datatype providing for Yes/No information, another
    datatype that allows additionally for a 'Maybe' value, and
    part of a simple "String" datatype where "Yes" happens to
    be the value of a fact expressed as a string. For the
    same datatype, unique values are enforced.

    The default fact data type of a fact value (unless specified
    otherwise) is the "String" datatype belonging to the Dingo default
    namespace.
    """

    value = models.TextField()
    title = models.CharField(max_length=256,
                             blank=True,
                             help_text="""A human-readable version of the value; useful for values
                                          that are part of standard vocabularies.""")
    description = models.TextField(blank=True,
                                   help_text="""A helptext/description of the value; useful for values that
                                                are part of standard vocabularies""")

    fact_data_type = models.ForeignKey("FactDataType",
                                       related_name = 'fact_value_set')




    storage_location = models.SmallIntegerField(choices=((dingos.DINGOS_VALUES_TABLE, "in FactValues table"),
                                                         (dingos.DINGOS_FILE_SYSTEM, "in Filesystem"),
                                                         (dingos.DINGOS_BLOB_TABLE, "in BLOB table" )),
                                                default=dingos.DINGOS_VALUES_TABLE,
                                                help_text="""Governs storage location of value""")


    class Meta:
        # The constraint below can cause problems if the entries in FactValue become to large:
        # Uniqueness forces the creation of a index on values
        # for testing uniqueness.
        unique_together = ('value', 'fact_data_type','storage_location')

    def __unicode__(self):
        return ("%s" % self.value)


dingos_class_map["FactValue"] = FactValue


class FactDataType(DingoModel):
    """
    Each fact value belongs to a fact data type. Fact data types
    are organized in name spaces -- a fact data type of a given
    name can only exist once in a given name space.
    """

    name = models.CharField(max_length=128,
                            help_text="Identifier for data type")

    description = models.TextField(blank=True)

    namespace = models.ForeignKey("DataTypeNameSpace",
                                  related_name = 'fact_data_type_set')

    class Meta:
        unique_together = ('name', 'namespace',)

    UNKNOWN_KIND = 0
    NO_VOCAB = 1
    VOCAB_SINGLE = 2
    VOCAB_MULTIPLE = 3
    REFERENCE = 4

    DATATYPE_KIND = ((UNKNOWN_KIND, "Unknown"),
                     (NO_VOCAB, "Not vocabulary!!"),
                     (VOCAB_SINGLE, "Vocabulary value (single choice)"),
                     (VOCAB_MULTIPLE, "Vocabulary value (multiple choice)"),
                     (REFERENCE, "Reference to InfoObject"),
    )

    kind = models.SmallIntegerField(choices=DATATYPE_KIND,
                                    default=UNKNOWN_KIND,
                                    help_text="""Governs, kind of data type.""")

    @staticmethod
    def autocomplete_search_fields():
        """ This is necessary for the autocomplete field offered by the Grappelli skin."""
        return ("name__icontains",)

    def __unicode__(self):
        ns_name = self.namespace.name
        if ns_name:
            return "%s (%s)" % (self.name, ns_name)
        else:
            return "%s" % self.name


dingos_class_map["FactDataType"] = FactDataType


class DataTypeNameSpace(DingoModel):
    """
    Fact data types are qualified by a namespace, allowing fact data types
    of the same name existing in parallel for different name spaces.

    We also use namespaces to denote owners/sources of information objects
    by giving objects an identifier consisting of a pair of (1) namespace
    of owner/source and (2) unique identifier. This follows the usage
    of "qualified names" as defined in the CybOx/STIX standards.
    """

    uri = models.URLField(max_length=255,
                          unique=True,
                          help_text="URI of namespace. Example: 'http://stix.mitre.org/default_vocabularies-1'")

    name = models.SlugField(max_length=50,
                            blank=True,
                            help_text="""Name of namespace. Example: 'cyboxVocabs'. This name may be used
                                         in XML output to denote the namespace. Note, however, that
                                         the defining characteristic of a namespace is the URI, not the
                                         name: the name is completely exchangeable.""")

    description = models.TextField(blank=True)

    def __unicode__(self):
        if self.name:
            return "%s (%s)" % (self.name, self.uri)
        else:
            return self.uri

    @staticmethod
    def autocomplete_search_fields():
        """ This is necessary for the autocomplete field offered by the Grappelli skin.
            See the Grappelli documenatation for details. """
        return ("name__icontains", "uri__icontains",)


dingos_class_map["DataTypeNameSpace"] = DataTypeNameSpace


class IdentifierNameSpace(DingoModel):
    """
    We use namespaces to denote owners/sources of information objects
    by giving objects an identifier consisting of a pair of (1) namespace
    of owner/source and (2) unique identifier. This follows the usage
    of "qualified names" as defined in the CybOx/STIX standards.
    """

    uri = models.URLField(max_length=255,
                          unique=True,
                          help_text="URI of namespace. Example: 'http://stix.mitre.org/default_vocabularies-1'")

    name = models.SlugField(max_length=50,
                            blank=True,
                            help_text="""Name of namespace. Example: 'my_organization'. This name may be used
                                         in XML output to denote the namespace. Note, however, that
                                         the defining characteristic of a namespace is the URI, not the
                                         name: the name is completely exchangeable.""")

    description = models.TextField(blank=True)

    def __unicode__(self):
        if self.name:
            return "%s (%s)" % (self.name, self.uri)
        else:
            return self.uri

    @staticmethod
    def autocomplete_search_fields():
        """ This is necessary for the autocomplete field offered by the Grappelli skin.
            See the Grappelli documenatation for details. """
        return ("name__icontains", "uri__icontains",)


dingos_class_map["IdentifierNameSpace"] = IdentifierNameSpace



class FactTerm(DingoModel):
    """
    Facts of any kind are essentially key-value pairs such
    as "First Name : Peter", "Gender: Male", etc. In Dingo,
    the values are managed with the FactValue model, whereas
    the keys are managed with the FactTerm model.

    We can provide structure to FactTerms by denoting
    them as "paths", e.g. "Person/Name/First", "Person/Name/Last",
    etc. In the case of XML import, we also treat attributes
    as fact terms: this is printed, e.g., as "Person@gender",
    but within the database, the attribute part of a FactTerm
    (if any) is stored in its own field.

    Here is an example of how we draw fact terms from
    XML documents. Consider the following piece of CybOx 1.0 XML::

        <FileObj:FileObjectType>
           <FileObj:File_Name datatype="String" condition="Equals">UNITED NATIONS COMPENSATION SCHEME...pdf
           </FileObj:File_Name>
           <FileObj:Hashes>
               <Common:Hash>
                   <Common:Type datatype="String">SHA256</Common:Type>
                   <Common:Simple_Hash_Value datatype="hexBinary" condition="Equals">
                   586fea79dd23a352a14c3f8bf3dbc9eb732e1d54f804a29160894aec55df4bd5
                   </Common:Simple_Hash_Value>
               </Common:Hash>
               <Common:Hash>
                   <Common:Type datatype="String">MD5</Common:Type>
                   <Common:Simple_Hash_Value datatype="hexBinary" condition="Equals">
                   471809c2092cecd633e43d465409a78c
                   </Common:Simple_Hash_Value>
               </Common:Hash>
           </FileObj:Hashes>
       </FileObj:FileObjectType>

    This gives rise to the following fact terms (the attributes regarding
    'datatype' are removed by the import, because we track this information
    in the FactDataType).

    - File_Name
    - File_Name@condition
    - Hashes/Hash/Type
    - Hashes/Hash/Simple_Hash_Value
    - Hashes/Hash/Simple_Hash_Value@condition


    By flattening the XML structure out into such fact terms, we would
    loose structural information, but we also keep
    track of tree structure information as we shall see later.

    The fact term model contains relatively little information.
    That  is, because the real information is contained in the "FactTerm2Type"
    model, that ties together FactTerms with all type information
    (FactDataType, InfoObjectFamily and InfoObjectType).
    """

    term = models.CharField(max_length=512,
                            help_text="""A path-like string such as "Header/Subject"
                                           or "Hashes/Hash/Simple_Hash_Value" """,
                            )

    attribute = models.CharField(max_length=128,
                                 help_text="""The key of an (XML) attribute that is part of the fact term (may be empty)""",
                                 )


    @staticmethod
    def autocomplete_search_fields():
        """ This is necessary for the autocomplete field offered by the Grappelli skin;
            it says that autocompletion is done on the 'term' field and checks whether
            what has been typed by the user appears anywhere in a known fact term
            (rather than, e.g., only completing terms that start with what the user has typed.)
            Do not forget to add the '@staticmethod' tag for this function if using
            the same trick for other models."""
        return ("term__icontains","attribute__icontains")

    def __unicode__(self):
        if self.attribute:
            return "%s@%s" % (self.term, self.attribute)
        else:
            return "%s" % self.term

    class Meta:
        unique_together = ('term', 'attribute')


dingos_class_map["FactTerm"] = FactTerm


class InfoObjectNaming(DingoModel):
    """
    DINGO provides a configurable naming system for Information Objects. How
    an Information Object is named is configured by associating it with one
    or more format strings that refer to facts within an object. The first
    format string for which all facts that it refers to are present
    in the format string is used for generating the name of an
    Information Object after import.
    """

    iobject_type = models.ForeignKey("InfoObjectType",
                                     related_name = 'iobject_type_set')
    format_string = models.TextField(help_text="""Format string for naming the information object. The format
                                                  string can refer to fact terms of facts that should be
                                                  present in an Information Object of the given type.""")
    position = models.PositiveSmallIntegerField("Position")

    class Meta:
        ordering = ['position']


dingos_class_map["InfoObjectNaming"] = InfoObjectNaming


class InfoObjectType(DingoModel):
    """
    Dingo organizes information in information objects. Each object
    is assigned an information-object type.

    Examples from the CybOx standard for information-object
    types:

    - File
    - Email
    - NetworkConnection
    - X509Certificate

    """


    name = models.SlugField(max_length=30,
                            help_text="""Name for a type of information object, e.g., 'Email',
                                         'File', 'Relationship', 'InvestigationStep' etc.""")

    description = models.TextField(blank=True)

    iobject_family = models.ForeignKey("InfoObjectFamily",
                                       related_name='iobject_type_set',
                                       help_text='Associated info-object family.')

    namespace = models.ForeignKey("DataTypeNamespace",
                                  blank=True,
                                  related_name='iobject_type_set',
                                  help_text='Namespace of information object type.')


    def __unicode__(self):
        return "%s:%s" % (self.iobject_family,self.name)

    @staticmethod
    def autocomplete_search_fields():
        """
        This is necessary for the autocomplete field offered by the Grappelli skin.
        """
        return ("name__icontains",)

    class Meta:
        unique_together = ('name', 'iobject_family', 'namespace')


dingos_class_map["InfoObjectType"] = InfoObjectType


class InfoObjectFamily(DingoModel):
    """
    There may be several source formats of information objects,
    for example:

    - CybOx
    - STIX
    - OpenIOC
    - ...

    The 'family' associated with an Information Objects informs
    about the source format.
    """

    name = models.SlugField(max_length=256,
                            unique=True,
                            help_text="Identifier for InfoObject Family")

    title = models.CharField(max_length=1024,
                             blank=True,
                             help_text="""A human-readable title for the InfoObject Family""")

    description = models.TextField(blank=True)


    @staticmethod
    def autocomplete_search_fields():
        """ This is necessary for the autocomplete field offered by
            the Grappelli skin."""
        return ("name__icontains",)

    def __unicode__(self):
        return "%s" % self.name


dingos_class_map["InfoObjectFamily"] = InfoObjectFamily


class Revision(DingoModel):
    """
    Both information-object families and information-object types may have
    different revisions (e.g., CybOx 1.0, CybOx 2.0, ...).
    We store revisions in a separate model.
    """
    name = models.CharField(max_length=32,
                            blank=True,
                            unique=True)

    @staticmethod
    def autocomplete_search_fields():
        """ This is necessary for the autocomplete field offered by
            the Grappelli skin."""
        return ("name__icontains",)

    def __unicode__(self):
        return "%s" % self.name


dingos_class_map["Revision"] = Revision


class FactTerm2Type(DingoModel):
    """
    A through model linking FactTerms, InfoObjectTypes and FactDataTypes: it
    specifies the data type of values of the fact term (FactDataType)
    for the given information-object type.

    For example::

       <URIObj:URIObjectType>
           <URIObj:Value datatype="AnyURI" condition="Equals">http://iprice.pl/image.php</URIObj:Value>
       </URIObj:URIObjectType>

    So, in URI-Objects, the fact-term "Value" has datatype "AnyURI". This example
    also shows, why we have to track the data type per object/iobject type:
    it stands to expect that "Value" may occur in other object types and may
    have other data types.
    """

    fact_term = models.ForeignKey("FactTerm",
                                  related_name='iobject_type_thru')

    iobject_type = models.ForeignKey("InfoObjectType",
                                     related_name='fact_term_thru')

    fact_data_types = models.ManyToManyField("FactDataType",
                                             related_name='fact_term_thru')

    title = models.CharField(max_length=1024,
                             blank=True,
                             help_text="""A human-readable title/summary of what the fact term describes for
                                          this InfoObjectType""")
    description = models.TextField(blank=True,
                                   help_text="""A comprehensive description of what the fact-term is used for
                                          in this InfoObjectType.""")

    class Meta:
        unique_together = ('iobject_type', 'fact_term')

    def __unicode__(self):
        return ("%s: %s x %s" % (
            self.iobject_type.iobject_family.name, self.fact_term.term, self.iobject_type.name))


dingos_class_map["FactTerm2Type"] = FactTerm2Type


class NodeID(DingoModel):
    """
    When flattening structured data such as XML files into facts (i.e., combinations
    of fact terms and fact values), we have to keep track of structural information.
    Consider the following example::

          <FileObj:FileObjectType>
           <FileObj:File_Name datatype="String" condition="Equals">UNITED NATIONS COMPENSATION SCHEME...pdf
           </FileObj:File_Name>
           <FileObj:Hashes>
               <Common:Hash>
                   <Common:Type datatype="String">SHA256</Common:Type>
                   <Common:Simple_Hash_Value datatype="hexBinary" condition="Equals">
                   586fea79dd23a352a14c3f8bf3dbc9eb732e1d54f804a29160894aec55df4bd5
                   </Common:Simple_Hash_Value>
               </Common:Hash>
               <Common:Hash>
                   <Common:Type datatype="String">MD5</Common:Type>
                   <Common:Simple_Hash_Value datatype="hexBinary" condition="Equals">
                   471809c2092cecd633e43d465409a78c
                   </Common:Simple_Hash_Value>
               </Common:Hash>
           </FileObj:Hashes>
       </FileObj:FileObjectType>

    This gives rise to the following node identifiers and associated facts,
    denoted in the following list by their fact terms:

    - 'N000': File_Name
    - 'N001:L000:N000' : Hashes/Hash/Type
    - 'N001:L000:N001': Hashes/Hash/Simple_Hash_Value
    - 'N001:L000:N001:A000' :Hashes/Hash/Simple_Hash_Value@condition
    - 'N001:L001:N000' : Hashes/Hash/Type
    - ...

    (The attributes 'datatype' are not considered, because that information is
    consumed during import for calculating datatype information and not duplicated
    in the information object's facts)


    The overall principle should be clear, but two special cases probably require
    additional explanation:

    - Lists such as the list of hash values, in which the same element name 'Hash'
      is repeated several times, are denoted with a 'L' in the node identifier.

    - Attributes are marked with a 'A' (we chose 'A' rather than '@' because then
      the lexicographic ordering on node identifier places attribute facts before
      the children of a node rather than behind the children.

    """
    name = models.CharField(max_length=255,
                            unique=True)

    def __unicode__(self):
        return "%s" % self.name


dingos_class_map["NodeID"] = NodeID


class InfoObject2Fact(DingoModel):
    """
    The model used for linking information objects and facts.
    It is a 'through' model which contains the following
    additional information:

    - node identifier: at which position does the fact
      occur in the information object?

    - if the fact is an attribute and their exists
      a fact to which this fact serves as attribute:
      a pointer to the latter.

      Note that there may be no such fact for an attribute.
      Consider the following example::

        <Cupboard>
          <Drawer items='2'>
            <Item color='silver'>Fork</Item>
            <Item color='black'>Knife></Item>
          </Drawer>
        </Cupboard>

      This is transformed into the following facts
      (with associated node identifier)

      - N000:N001:A000: Cupboard/Drawer@items = 2
      - N000:N001:L000: Cupboard/Drawer/Item = Fork
      - N000:N001:L000:A000 Cupboard/Drawer/Item@color = silver
      - N000:N001:L001: Cupboard/Drawer/Item = Knife
      - N000:N001:L001:A000 Cupboard/Drawer/Item@color = black

      Here, for the fact with node id 'N000:N001:A000' there
      is no associated fact to which '2' is the attribute with
      key 'items': we have no fact 'Cupboard/Drawer', because
      we have flattened the tree into facts and 'Cupboard/Drawer'
      itself is not associated with a value but provides
      hierarchical information.

      For the color attributes 'Cupboards/Drawer/Item@color',
      however, there exist the two facts 'Cupboards/Drawer/Item'
      and DINGO will fill in the information in 'attributed_fact'.

      Note that the reverse relation of 'attributed_fact'
      is named 'attributes', i.e., for the fact within an
      Information Object, you can see whether it has a direct
      attribute.

    """

    iobject = models.ForeignKey("InfoObject",
                                related_name="fact_thru",
                                )

    fact = models.ForeignKey("Fact",
                             related_name="iobject_thru",
                             )

    attributed_fact = models.ForeignKey("InfoObject2Fact",
                                        null=True,
                                        related_name="attributes")

    node_id = models.ForeignKey("NodeID")



    @property
    def marking_thru(self):
        """
        Return the back-pointer to  markings that
        may have been attached via Django's content type mechanism.
        """
        self_django_type = ContentType.objects.get_for_model(self)
        return Marking2X.objects.filter(content_type__pk=self_django_type.id,
                                        object_id=self.id)


    def __unicode__(self):
        return self.fact.fact_term.attribute

    class Meta:
        # This ordering is crucial for proper display of facts
        # in the detail view of an Information Object
        ordering = ['node_id__name']


dingos_class_map["InfoObject2Fact"] = InfoObject2Fact


class Fact(DingoModel):
    """
    In facts, we associate fact terms with one or more fact values.

    A fact may contain a (single) reference to another
    information object. Note that the reference occurs via the
    identifier rather than the information object directly: the semantics
    is that always the latest revision is referred to. If the reference
    to a specific revision is required, the time stamp of that
    revision must also be specified.

    Note that facts do not have a direct foreign key
    relationship with information objects: we link
    information objects and facts via a thru model,
    namely InfoObject2Fact.
    """

    fact_term = models.ForeignKey(FactTerm,
                                  help_text="Pointer to fact term described by this fact.",
                                  )
    fact_values = models.ManyToManyField("FactValue",
                                         null=True,
                                         help_text="""Value(s) of that fact""")

    value_iobject_id = models.ForeignKey("Identifier",
                                         related_name="value_of_set",
                                         null=True,
                                         blank=True,
                                         help_text="""As alternative to a text-based value stored in a fact,
                                                       an iobject can be linked. In this case, there should
                                                       be no fact values associated with the fact.""")

    value_iobject_ts = models.DateTimeField(null=True,
                                            help_text="""Used to reference a specific revision of an information
                                                         object rather than the latest revision.""")


    class Meta:
        # Here, we cannot have database-enforced uniqueness, because we need
        # uniqueness regarding also M2M-related objects, namely the fact values.
        # Refer to the code in get_or_create_fact that shows how to query whether
        # such a fact may already exist.
        pass

    def __unicode__(self):
        fact_values = self.fact_values.all()
        if fact_values.count() > 1:
            more_values = "(%s more values)" % (fact_values.count()-1)
        else:
            more_values = ""
        if len(fact_values[0].value)>50:
            longer = "..."
        else:
            longer = ''
        return "%s: %s%s %s" % (self.fact_term.term, self.fact_values.all()[0].value[0:50],longer,more_values)


    @staticmethod
    def autocomplete_search_fields():
        """ This is necessary for the autocomplete field offered by the Grappelli skin."""
        return ("fact_term__name__icontains",)


dingos_class_map["Fact"] = Fact


class InfoObject(DingoModel):
    """
    So this is the heart of Dingo, the information object class:

    an information object is of a certain type and belongs to
    a certain family. To give examples from CybOx: the type may
    be 'File' and the family is, well, CybOx. For both
    type and class, we may also record a revision (e.g.
    file object of definition "1_2" and family CybOx 2.0).

    An information object contains information via a mapping
    from the information object to facts.

    An information object is marked with an identifier
    and a timestamp. For the same identifier there may be
    several timestamps, thus recording different versions
    of the information object. The newest version of
    an information object can always be found via
    the identifier model: 'self.identifier.latest' always
    points to the latest version.

    If a given information object represents the latest
    revision, the method 'latest_of' (provided as
    reverse relation-ship from the Identifier model)
    is not None but contains 'self' as only element.

    An information object has a name, which is derived
    from the facts contained in the object: the definition
    of how to derive this name is contained in the
    field 'naming_scheme' of the information-object type
    to which the information object belongs. In the interest
    of efficiency when displaying lists of objects,
    the name is not calculated on the fly but should
    be calculated and stored after the information object
    has been created or updated.
    """

    identifier = models.ForeignKey("Identifier",
                                   related_name = 'iobject_set'
                                  )

    timestamp = models.DateTimeField()

    create_timestamp = models.DateTimeField()

    facts = models.ManyToManyField("Fact",
                                   through="InfoObject2Fact",
                                   help_text="Facts that are content of this enrichment")

    iobject_type = models.ForeignKey(InfoObjectType,
                                     related_name = 'iobject_set',
                                     help_text="Each enrichment has an information object type.")

    iobject_type_revision = models.ForeignKey(Revision,
                                              help_text="Each enrichment has an information object type.",
                                              related_name="+")

    iobject_family = models.ForeignKey("InfoObjectFamily",
                                       related_name = 'iobject_set',
                                       help_text="Type of enrichment, e.g. 'CYBOX'")

    iobject_family_revision = models.ForeignKey("Revision",
                                                help_text="Revision of enrichment type , e.g. '1.0'",
                                                related_name='+')
    uri = models.URLField(blank=True,
                          help_text="""URI pointing to further
                                       information concerning this
                                       enrichment, e.g., the HTML
                                       report of a malware analysis
                                       through Cuckoo or similar.""")

    name = models.CharField(max_length=255,
                            blank=True,
                            default='Unnamed',
                            editable=False,
                            help_text="""Name of the information object, usually auto generated.
                                         from type and facts flagged as 'naming'.""")

    @property
    def marking_thru(self):
        """
        Return the back-pointer to a marking that may
        has been attached via Django's content type mechanism.
        """

        self_django_type = ContentType.objects.get_for_model(self)
        return Marking2X.objects.filter(content_type__pk=self_django_type.id,
                                        object_id=self.id)


    class Meta:
        unique_together = (('identifier', 'timestamp'),)
        ordering = ['-timestamp']


    def __unicode__(self):
        return "%s: %s" % (self.iobject_type, self.name)

    def is_empty(self):
        """
        Returns true if the enrichment is empty

        :return:
        """

        return self.fact_thru.count() == 0

    @property
    def embedded_in(self):
        """
        Used in the view for the InfoObject (in order to be able to use the standard class-based object view.
        Should be removed from here and put into a proper custom view for the object.

        This query only returns embedding objects of the latest revision: to change
        this, the filter 'iobject__timestamp=F('iobject__identifier__latest__timestamp' must
        be removed.
        """

        return self._DCM['InfoObject2Fact']. \
            objects. \
            filter(fact__value_iobject_id=self.identifier). \
            filter(iobject__timestamp=F('iobject__identifier__latest__timestamp')). \
            order_by('-iobject__timestamp') \
            .values_list(
            'iobject',
            'iobject__identifier__namespace__uri',
            'iobject__identifier__uid',
            'iobject__timestamp',
            'iobject__name',
            'fact__value_iobject_ts',
            'fact__fact_term__term', 
            'node_id__name').distinct()

    def add_fact(self,
                 fact_term_name,
                 fact_term_attribute,
                 fact_dt_name='String',
                 fact_dt_namespace_name=None,
                 fact_dt_namespace_uri=DINGOS_NAMESPACE_URI,
                 fact_dt_kind=FactDataType.UNKNOWN_KIND,
                 values=None,
                 value_iobject_id=None,
                 value_iobject_ts=None,
                 node_id_name='',
                 is_attribute=False):
        """
        Add a fact to the iobject. If a fact term for the iobject type
        with the given fact data type and source (CYBOX, etc.) does not
        exist yet, it is created.

        """

        if not values:
            values = []

        # get or create fact_term
        fact_term, created = get_or_create_fact_term(iobject_family_name=self.iobject_family.name,
                                                     fact_term_name=fact_term_name,
                                                     fact_term_attribute=fact_term_attribute,
                                                     iobject_type_name=self.iobject_type.name,
                                                     iobject_type_namespace_uri=self.iobject_type.namespace.uri,
                                                     fact_dt_name=fact_dt_name,
                                                     fact_dt_kind=fact_dt_kind,
                                                     fact_dt_namespace_name=fact_dt_namespace_name,
                                                     fact_dt_namespace_uri=fact_dt_namespace_uri,
                                                     dingos_class_map=self._DCM)


        # get or create fact object

        fact_obj, created = get_or_create_fact(fact_term,
                                               fact_dt_name=fact_dt_name,
                                               fact_dt_namespace_uri=fact_dt_namespace_uri,
                                               values=values,
                                               value_iobject_id=value_iobject_id,
                                               value_iobject_ts=value_iobject_ts,
                                               )


        # get or create node identifier

        node_id, created = dingos_class_map['NodeID'].objects.get_or_create(name=node_id_name)

        # If this is an attribute: determine whether there is a fact for which this fact is
        # an attribute.

        node_id_name_components = node_id_name.split(':')

        attributed_io2f = None
        if node_id_name_components[-1] and node_id_name_components[-1][0] == 'A':
            attributed_node_id_name = ":".join(node_id_name_components[:-1])

            try:
                attributed_io2f = dingos_class_map['InfoObject2Fact'].objects.get(iobject=self,
                                                                                 node_id__name=attributed_node_id_name)
            except InfoObject2Fact.DoesNotExist:
                pass

        e2f = self._DCM['InfoObject2Fact'].objects.create(
            node_id=node_id,
            iobject=self,
            fact=fact_obj,
            attributed_fact=attributed_io2f)

        return e2f

    def from_dict(self,
                  dingos_obj_dict,
                  config_hooks=None,
                  namespace_dict=None,
                  special_ft_handler=None):
        """
        Convert DingoObjDict to facts and associate resulting facts with this information object.
        """


        # Instantiate default parameters

        if not config_hooks:
            config_hooks = {}

        datatype_extractor = config_hooks.get('datatype_extractor', (lambda io, f, i, n, d: False))

        special_ft_handler = config_hooks.get('special_ft_handler', None)

        attr_ignore_predicate = config_hooks.get('attr_ignore_predicate', None)

        force_nonleaf_fact_predicate = config_hooks.get('force_nonleaf_fact_predicate', None)

        if not namespace_dict:
            namespace_dict = {}

        if not self.is_empty():
            logger.error("Attempt to import a dictionary into a non-empty info object")
            return

        # Flatten the DingoObjDict

        (flat_list, attrs) = dingos_obj_dict.flatten(attr_ignore_predicate=attr_ignore_predicate,
                                                     force_nonleaf_fact_predicate=force_nonleaf_fact_predicate)

        for fact in flat_list:

            # Collect the information about all attributes relevant
            # for this node (i.e., they occur either directly on
            # the node or on an ancestor node).

            # The following code was used to generate a attribute dictionary::
            #
            #     attr_info = ExtendedSortedDict()
            #     for attr_node in attrs.keys():
            #         if fact['node_id'].startswith(attr_node):
            #            for (key, value) in attrs[attr_node].items():
            #                attr_info.chained_set(value, 'set', key, attr_node)
            #
            # The dictionary contained full information about also the attributes given
            # to ancestor nodes. So far, we have not required this
            # information for our imports, and thus simplify to a
            # dictionary that only contains the attributes directly
            # associated with the current node.
            #
            # If, at a later stage, we find we need this kind of information,
            # we will can use Django's MultiValueDict to add additional
            # information without changing the signature of the receiving
            # predicate and handler functions.


            attr_info = dict(attrs.get(fact['node_id'],[]))

            #for attr_node in attrs.keys():
            #    if fact['node_id'] == (attr_node):
            #        for (key, value) in attrs[attr_node].items():
            #            attr_info.chained_set(value, 'set', key, attr_node)


            # Fill dictionary with arguments for call to 'add_fact'

            add_fact_kargs = {}
            add_fact_kargs['fact_dt_kind'] = FactDataType.UNKNOWN_KIND
            add_fact_kargs['fact_dt_namespace_name'] = "%s-%s" % (
                self.iobject_family.name, self.iobject_family_revision.name)

            # See whether the datatype extractor has found a datatype for the value

            datatype_found = datatype_extractor(self, fact, attr_info, namespace_dict, add_fact_kargs)

            if not datatype_found:
                add_fact_kargs = {}
                add_fact_kargs['fact_dt_kind'] = FactDataType.NO_VOCAB
                add_fact_kargs['fact_dt_namespace_name'] = DINGOS_NAMESPACE_SLUG
                add_fact_kargs['fact_dt_namespace_uri'] = DINGOS_NAMESPACE_URI
            else:
                # Check whether the datatype extractor added namespace information
                # If not, add some here
                if not 'fact_dt_namespace_uri' in add_fact_kargs:
                    add_fact_kargs['fact_dt_namespace_uri'] = namespace_dict.get(
                        add_fact_kargs['fact_dt_namespace_name'], '%s/%s' % (
                            DINGOS_NAMESPACE_URI, self.iobject_family))

            add_fact_kargs['fact_term_name'] = fact['term']
            add_fact_kargs['fact_term_attribute'] = fact['attribute']
            add_fact_kargs['values'] = [fact['value']]
            add_fact_kargs['node_id_name'] = fact['node_id']

            handler_return_value = True

            logger.debug("Treating fact (before special handler list) %s with attr_info %s and kargs %s" % (fact, attr_info, add_fact_kargs))
            # Below, go through the handlers in the special_ft_handler list --
            # if the predicate returns True for the fact, execute the handler
            # on the fact. If a handler returns False/None, the fact is *not*
            # added. This should only be done, if the handler has added the
            # fact -- otherwise the sequence of node identifiers is messed up!

            if special_ft_handler:
                for (predicate, handler) in special_ft_handler:
                    if predicate(fact, attr_info):
                        handler_return_value = handler(self, fact, attr_info, add_fact_kargs)
                        if not handler_return_value:
                            break
            logger.debug("Treating fact (before special handler list) %s with attr_info %s and kargs %s" % (fact, attr_info, add_fact_kargs))
            if (handler_return_value == True):
                e2f_obj = self.add_fact(**add_fact_kargs)
            elif not handler_return_value:
                continue
            else:
                e2f_obj = handler_return_value
        self.set_name()


    def to_dict(self,include_node_id=False):
        flat_result = []

        fact_thrus = self.fact_thru.all().prefetch_related(
                                                           'fact__fact_term',
                                                           'fact__fact_values',
                                                           'fact__fact_values__fact_data_type',
                                                           'fact__fact_values__fact_data_type__namespace',
                                                           'fact__value_iobject_id',
                                                           'fact__value_iobject_id__namespace',
                                                           'node_id')

        #fact_thrus = self.fact_thru.all()
        for fact_thru in fact_thrus:
            value_list = []
            first = True
            fact_datatype_name = None
            fact_datatype_ns = None
            for fact_value in fact_thru.fact.fact_values.all():
                if first:
                    fact_datatype_name = fact_value.fact_data_type.name
                    fact_datatype_ns = fact_value.fact_data_type.namespace.uri
                    if (fact_datatype_name == DINGOS_DEFAULT_FACT_DATATYPE and
                        fact_datatype_ns == DINGOS_NAMESPACE_URI):

                        fact_datatype_name = None
                        fact_datatype_ns = None
                        first = False
                value_list.append(fact_value.value)

            fact_dict = {'node_id': fact_thru.node_id.name,
                         'term': fact_thru.fact.fact_term.term,
                         'attribute' : fact_thru.fact.fact_term.attribute,
                         '@@type': fact_datatype_name,
                         '@@type_ns': fact_datatype_ns,
                         'value_list': value_list}

            if fact_thru.fact.value_iobject_id:
                value_iobject_id_ns = fact_thru.fact.value_iobject_id.namespace.uri
                value_iobject_id  =fact_thru.fact.value_iobject_id.uid
                fact_dict['@@idref_ns'] = fact_thru.fact.value_iobject_id.namespace.uri
                fact_dict['@@idref_id'] = fact_thru.fact.value_iobject_id.uid
            flat_result.append(fact_dict)

        result = DingoObjDict()
        result.from_flat_repr(flat_result,include_node_id=include_node_id)
        result['@@iobject_type'] = self.iobject_type.name
        result['@@iobject_type_ns'] = self.iobject_type.namespace.uri
        #result = result.to_tuple()
        return result

    def show_fact_terms(self,level):
        """Returns a list of the fact terms (split in 'term' and 'attribute') that
         occur in this InfoObject."""

        fact_terms = self._DCM["FactTerm"].objects.filter(fact__infoobject__pk=self.pk). \
            distinct('term','attribute').values('term','attribute')


        #fact_terms = self._DCM["FactTerm"].objects.filter(fact__iobject_thru__iobject__pk=self.pk).\
        #    filter(fact__iobject_thru__node_id__name__startswith=level).order_by('term','attribute').\
        #    distinct('term','attribute').values_list('fact__iobject_thru__node_id__name','term','attribute')

        return list(fact_terms)

    def extract_name(self):
        """
        We want to dynamically name the iobject based on its type and
        selected facts. This is, what this function does: it collects
        all facts that have "naming" set and concatenates them in the
        order given by the "naming" argument.
        If no fact has such a flag, the type is used

        """
        name = None

        name_schemas = self._DCM["InfoObjectNaming"].objects.filter(iobject_type=self.iobject_type).order_by(
            'position').values_list('format_string', flat=True)

        name_components = []

        # We retrieve all facts of the object

        fact_list = self._DCM['Fact'].objects.filter(iobject_thru__iobject=self).order_by(
            'iobject_thru__node_id__name').values_list('iobject_thru__node_id__name',
                                                        'fact_term__term',
                                                        'fact_term__attribute',
                                                        'fact_values__value',
                                                        'value_iobject_id__latest__name')

        # We build a dictionary that will then be used for the format string

        fact_dict = {}
        counter = 0
        for (node_id, fact_term, attribute, value, related_obj_name) in fact_list:
            #print fact_list[counter]
            if type(value)==type([]):
                value = "%s,..." % value
            if related_obj_name:
                value = related_obj_name

            if attribute:
                fact_term = "%s@%s" % (fact_term, attribute)

            if not fact_term in fact_dict:
                fact_dict[fact_term] = value

            # In addition to the mapping of fact term to (first) value,
            # we add a way to refer to fact terms and values by their
            # node id

            fact_dict["term_of_node_%s" % node_id] = fact_term
            fact_dict["value_of_node_%s" % node_id] = value
            if counter < 10:
                fact_dict["term_of_fact_num_%01d" % counter] = fact_term
                fact_dict["value_of_fact_num_%01d" % counter] = value
            counter += 1
        fact_dict["fact_count_equal_%s?" % counter] = ""
        fact_dict["fact_count"] = "%s" % counter

        name_found = False

        # Now, we go through the format strings and see whether one of them
        # is successfully instantiated via the fact dictionary
        #print fact_dict

        for format_string in name_schemas:
            # Massage the format string such that we can use it as Python format string

            # Escape possible '%'
            format_string = format_string.replace('%','\%')
            # Replace placeholder definitions with python string formatting
            format_string = RE_SEARCH_PLACEHOLDERS.sub("%(\\1)s", format_string)




            try:
                name = format_string % fact_dict
                name_found = True
                break
            except:
                continue

        if not name_found:
            if self.iobject_type.name == 'PLACEHOLDER':
                return "PLACEHOLDER"
            else:
                return "%s (%s facts)" % (self.iobject_type.name,counter)
        else:
            return name

    def set_name(self,name=None):
        """
        Set the name of the object. If no name is given, the
        name is extracted via the extract_name method.
        """
        if name:
            self.name = name[:254]
        else:
            self.name = self.extract_name()[:254]

        self.save()

        return self.name

    def add_relation(self,
                     target_id=None,
                     relation_types=None,
                     fact_dt_namespace_name=None,
                     fact_dt_namespace_uri=DINGOS_NAMESPACE_URI,
                     fact_dt_kind=FactDataType.UNKNOWN_KIND,
                     fact_dt_name='String',
                     metadata_dict=None,
                     markings=None
    ):
        """
        Add a relationship between this object and another object.
        """
        if not markings:
            markings = []

        if relation_types == None:
            relation_types = []

        # Create fact-term for relation types
        relation_type_ft, created = get_or_create_fact_term(iobject_family_name=self.iobject_family.name,
                                                            fact_term_name=DINGOS_RELATION_TYPE_FACTTERM_NAME,
                                                            iobject_type_name=self.iobject_type.name,
                                                            iobject_type_namespace_uri=self.iobject_type.namespace.uri,
                                                            fact_dt_name=fact_dt_name,
                                                            fact_dt_namespace_name=fact_dt_namespace_name,
                                                            fact_dt_kind=fact_dt_kind,
                                                            fact_dt_namespace_uri=fact_dt_namespace_uri)

        # Create fact containing relation types
        relation_type_fact, created = get_or_create_fact(fact_term=relation_type_ft,
                                                         fact_dt_name=fact_dt_name,
                                                         fact_dt_namespace_uri=fact_dt_namespace_uri,
                                                         values=relation_types,
                                                         value_iobject_id=None,
                                                         value_iobject_ts=None,
                                                         )

        rel_target_id = target_id
        rel_source_id = self.identifier

        # Create relation object
        relation, created = self._DCM['Relation'].objects.get_or_create(
            source_id=rel_source_id,
            target_id=rel_target_id,
            relation_type=relation_type_fact)

        # Add markings
        for marking in markings:
            Marking2X.objects.create(marked=relation,
                                     marking=marking)

        if metadata_dict:
            # If the relation already existed and had associated metadata,
            # we retrieve the identifier of that metadata object and
            # write the current metadata as new revision. Otherwise,
            # we create a new identifier.

            if relation.metadata_id:
                rel_identifier_uid = relation.metadata_id.uid
                rel_identifier_namespace_uri = relation.metadata_id.namespace.uri
            else:
                rel_identifier_uid = None
                rel_identifier_namespace_uri = DINGOS_ID_NAMESPACE_URI

            metadata_iobject, created = get_or_create_iobject(identifier_uid=rel_identifier_uid,
                                                              identifier_namespace_uri=rel_identifier_namespace_uri,
                                                              iobject_type_name=DINGOS_RELATION_METADATA_OBJECT_TYPE_NAME,
                                                              iobject_type_namespace_uri=DINGOS_NAMESPACE_URI,
                                                              iobject_type_revision_name=DINGOS_REVISION_NAME,
                                                              iobject_family_name=DINGOS_IOBJECT_FAMILY_NAME,
                                                              iobject_family_revision_name=DINGOS_REVISION_NAME,
                                                              timestamp=None,
                                                              overwrite=False)
            metadata_iobject.from_dict(metadata_dict)

        return relation

    @property
    def view_link(self):
        """
        Return link to Dingo's Detail View for InfoObjects. We use this
        in the Django Admin interface to provide a link to this page
        whenever an information object is shown (e.g., in the InfoObject
        list view).
        """
        iobject_url = urlresolvers.reverse('url.dingos.view.infoobject', args=(self.id,))
        return mark_safe(
            """<a href="%s"><img src="/static/admin/img/selector-search.gif" alt="Lookup" height="16" width="16">""" % iobject_url)


dingos_class_map["InfoObject"] = InfoObject


class Identifier(DingoModel):
    """
    Each information object has an identifier, that consists of a name space,
    denoting the owner/source of an object, and a uid. The 'latest' field
    should always point to the most recent information object of a given
    identifier, i.e., the one with the most recent time stamp.
    """

    uid = models.SlugField(max_length=255)
    namespace = models.ForeignKey("IdentifierNameSpace")

    latest = models.OneToOneField(InfoObject,
                                  null=True, # need this for creation (hen-egg problem)
                                  related_name="latest_of")


    def __unicode__(self):
        return "%s (%s)" % (self.uid, self.namespace.uri)

    class Meta:
        unique_together = ('uid', 'namespace',)


dingos_class_map["Identifier"] = Identifier


class Relation(DingoModel):
    """
    Relations between information objects.

    """

    source_id = models.ForeignKey(Identifier,
                                  null=True,
                                  related_name='yields_via',
                                  help_text="""Pointer to source iobject, i.e., the iobject from
                                               which something was derived""")



    target_id = models.ForeignKey(Identifier,
                                  null=True,
                                  related_name='yielded_by_via',
                                  help_text="Pointer to derived iobject")


    relation_type = models.ForeignKey(Fact,
                                      help_text="Description of nature of relation in direction source to target.")

    metadata_id = models.ForeignKey(Identifier,
                                    null=True,
                                    related_name='+',
                                    help_text="InfoObject containing metadata about relation.")

    @property
    def marking_thru(self):
        """
        Return the back-pointer to markings that may have
        been attached via Django's content type mechanism.
        """

        self_django_type = ContentType.objects.get_for_model(self)
        return Marking2X.objects.filter(content_type__pk=self_django_type.id,
                                        object_id=self.id)


    class Meta:
        unique_together = ("source_id",
                           "target_id",
                           "relation_type")

    def __unicode__(self):
        return "Relation %s" % self.pk


dingos_class_map["Relation"] = Relation

class BlobStorage(models.Model):
    """
    A table for storing large values.
    """

    sha256 = models.CharField(unique=True,
                              max_length=64
                              )

    content = models.TextField(blank=True)


dingos_class_map["BlobStorage"] = BlobStorage

class Marking2X(models.Model):
    """
    Information Objects, single facts within an information object, or Relations
    can be marked with another information object. We use Django's content-type
    mechanism to allow relationships with several models.
    """
    marking = models.ForeignKey(InfoObject,
                                related_name='marked_item_thru')
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    marked = generic.GenericForeignKey('content_type', 'object_id')


dingos_class_map["Marking2X"] = Marking2X


def get_or_create_iobject(identifier_uid,
                          identifier_namespace_uri,
                          iobject_type_name,
                          iobject_type_namespace_uri,
                          iobject_type_revision_name,
                          iobject_family_name,
                          iobject_family_revision_name="",
                          identifier_namespace_name="",
                          timestamp=None,
                          create_timestamp=None,
                          overwrite=False,
                          dingos_class_map=dingos_class_map):
    """
    Get or create an information object.
    """

    # create or retrieve the iobject type and revision

    # create or retrieve identifier

    id_namespace, created = dingos_class_map['IdentifierNameSpace'].objects.get_or_create(uri=identifier_namespace_uri)

    if created and identifier_namespace_name:
        id_namespace.name = identifier_namespace_name
        id_namespace.save()

    identifier, created = dingos_class_map['Identifier'].objects.get_or_create(uid=identifier_uid,
                                                                              namespace=id_namespace,
                                                                              defaults={'latest': None})

    iobject_type_namespace, created = dingos_class_map['DataTypeNameSpace'].objects.get_or_create(uri=iobject_type_namespace_uri)

    iobject_family, created = dingos_class_map['InfoObjectFamily'].objects.get_or_create(name=iobject_family_name)
    iobject_family_revision, created = dingos_class_map['Revision'].objects.get_or_create(
        name=iobject_family_revision_name)

    # create or retrieve the iobject type
    iobject_type, created = dingos_class_map['InfoObjectType'].objects.get_or_create(name=iobject_type_name,
                                                                                    iobject_family=iobject_family,
                                                                                    namespace=iobject_type_namespace)
    iobject_type_revision, created = dingos_class_map['Revision'].objects.get_or_create(name=iobject_type_revision_name)

    if not create_timestamp:
        create_timestamp = timezone.now()
    if not timestamp:
        timestamp = create_timestamp

    if overwrite:
        iobject = overwrite
        created = False


    else:
        iobject, created = dingos_class_map["InfoObject"].objects.get_or_create(identifier=identifier,
                                                                               timestamp=timestamp,
                                                                               defaults={'iobject_family': iobject_family,
                                                                                         'iobject_family_revision': iobject_family_revision,
                                                                                         'iobject_type': iobject_type,
                                                                                         'iobject_type_revision': iobject_type_revision,
                                                                                         'create_timestamp': create_timestamp})
    if created:
        iobject.set_name()
        iobject.save()

    elif overwrite:
        iobject.timestamp = timestamp
        iobject.create_timestamp = create_timestamp
        iobject.iobject_family = iobject_family
        iobject.iobject_family_revision = iobject_family_revision
        iobject.iobject_type = iobject_type
        iobject.iobject_type_revision = iobject_type_revision
        iobject.set_name()
        iobject.save()

    logger.debug(
        "Created iobject with %s (created was %s) and %s (overwrite %s)" % (iobject.identifier, timestamp, created, overwrite))
    return iobject, created


def get_or_create_fact(fact_term,
                       fact_dt_name='String',
                       fact_dt_namespace_uri=DINGOS_NAMESPACE_URI,
                       values=None,
                       value_iobject_id=None,
                       value_iobject_ts=None,
                       ):
    """
    Get or create a fact object.
    """

    if not values:
        values = []


    vocab_namespace, created = dingos_class_map['DataTypeNameSpace'].objects.get_or_create(uri=fact_dt_namespace_uri)

    fact_data_type, created = dingos_class_map['FactDataType'].objects.get_or_create(name=fact_dt_name,
                                                                                    namespace=vocab_namespace)

    # Maybe we already have a fact with exactly the same fact term and the same fact values?
    # We start by looking at the number of values

    value_objects = []

    for value in values:
        storage_location=dingos.DINGOS_VALUES_TABLE
        # collect (create or get) the required value objects
        if value == None:
            value = ''
        if isinstance(value,tuple):
            # If a value is wrapped in a tuple, the second component of the tuple
            # specifies the storage location of the value.
            value, storage_location = value

        if storage_location == dingos.DINGOS_VALUES_TABLE:
            # If the value is larger than a given size, the value is written to disk, instead.
            # We use this to keep too large values out of the database. Depending on how the
            # database is set up, this may be necessary to allow indexing, which in turn is
            # required to check uniqueness on values.

            if len(value) > dingos.DINGOS_MAX_VALUE_SIZE_WRITTEN_TO_VALUE_TABLE:
                (value_hash,storage_location) = write_large_value(value)
                value = value_hash


        fact_value, created = dingos_class_map['FactValue'].objects.get_or_create(value=value,
                                                                                 fact_data_type=fact_data_type,
                                                                                 storage_location=storage_location)
        value_objects.append(fact_value)



    # Do we already have a fact with given fact term and given values?
    #
    # For understanding the query below better, see https://groups.google.com/forum/#!topic/django-users/X9TCSrBn57Y.
    # The double query is necessary, because the first count counts the number of selected
    # fact_value objects, not the number of total objects for each fact.

    matching_facts = Fact.objects.filter(fact_values__in=value_objects). \
        annotate(num_values=Count('fact_values')). \
        filter(num_values=len(value_objects)). \
        filter(value_iobject_id=value_iobject_id). \
        filter(value_iobject_ts=value_iobject_ts). \
        filter(fact_term=fact_term). \
        exclude(id__in= \
        Fact.objects.annotate(total_values=Count('fact_values')). \
            filter(total_values__gt=len(value_objects)))

    created = True
    try:
        fact_obj = matching_facts[0]
        created = False
        logger.debug("FOUND MATCHING OBJECT with pk %s" % fact_obj.pk)
    except:
        fact_obj = dingos_class_map['Fact'].objects.create(fact_term=fact_term,
                                                          value_iobject_id=value_iobject_id,
                                                          value_iobject_ts=value_iobject_ts,
                                                           )

        fact_obj.fact_values.add(*value_objects)
        fact_obj.save()


    return fact_obj, created


def get_or_create_fact_term(iobject_family_name,
                            fact_term_name,
                            fact_term_attribute,
                            iobject_type_name,
                            iobject_type_namespace_uri,
                            fact_dt_name=DINGOS_DEFAULT_FACT_DATATYPE,
                            fact_dt_namespace_name=None,
                            fact_dt_kind=FactDataType.UNKNOWN_KIND,
                            fact_dt_namespace_uri=DINGOS_NAMESPACE_URI,
                            dingos_class_map=dingos_class_map
):
    """
    Get or create a fact term.
    """

    if not fact_term_attribute:
        fact_term_attribute = ''

    # create or retrieve the enrichment type and revision

    iobject_family, created = dingos_class_map['InfoObjectFamily'].objects.get_or_create(name=iobject_family_name)

    # create or retrieve namespace of data type

    fact_dt_namespace, created = dingos_class_map['DataTypeNameSpace'].objects.get_or_create(uri=fact_dt_namespace_uri)

    # create or retrieve namespace of the infoobject type

    iobject_type_namespace, created = dingos_class_map['DataTypeNameSpace'].objects.get_or_create(uri=iobject_type_namespace_uri)

    if created and fact_dt_namespace_name:
        fact_dt_namespace.name = fact_dt_namespace_name
        fact_dt_namespace.save()


    # create or retrieve the fact-value data type object
    fact_dt, created = dingos_class_map['FactDataType'].objects.get_or_create(name=fact_dt_name,
                                                                              namespace=fact_dt_namespace)

    if created:
        fact_dt.kind = fact_dt_kind
        fact_dt.save()

    # create or retreive the iobject type
    iobject_type, created = dingos_class_map['InfoObjectType'].objects.get_or_create(name=iobject_type_name,
                                                                                    iobject_family=iobject_family,
                                                                                    namespace=iobject_type_namespace)

    fact_term, created = dingos_class_map['FactTerm'].objects.get_or_create(term=fact_term_name,
                                                                           attribute=fact_term_attribute)

    fact_term_2_type, dummy = dingos_class_map['FactTerm2Type'].objects.get_or_create(fact_term=fact_term,
                                                                                     iobject_type=iobject_type,
                                                                                     )

    fact_term_2_type.fact_data_types.add(fact_dt)

    fact_term_2_type.save()

    return fact_term, created


def write_large_value(value,storage_location=dingos.DINGOS_LARGE_VALUE_DESTINATION):
    value_hash = hashlib.sha256(value).hexdigest()
    if storage_location == dingos.DINGOS_FILE_SYSTEM:

        file_name = '%s.blob' % (value_hash)

        if dingos.DINGOS_BLOB_STORAGE.exists(file_name):
            dingos.DINGOS_BLOB_STORAGE.delete(file_name)

        dingos.DINGOS_BLOB_STORAGE.save(file_name, ContentFile(value))

    else:
        storage_location = dingos.DINGOS_BLOB_TABLE
        # The blob storage table is default
        dingos_class_map['BlobStorage'].objects.get_or_create(sha256=value_hash,
                                                              content=value)
    return (value_hash,storage_location)




# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import sys
import pprint
import json
import re

from django.core.management.base import BaseCommand, CommandError
from optparse import make_option

from dingos.models import InfoObjectType, InfoObjectNaming, DataTypeNameSpace, InfoObjectFamily

pp = pprint.PrettyPrinter(indent=2)

class Command(BaseCommand):
    """

    """
    args = ''
    help = 'Export and import of naming schemas'

    option_list = BaseCommand.option_list + (
        make_option('-o', '--output-file',
                    action='store',
                    dest='output_file',
                    default=None,
                    help='File to which naming info is to be written. If this parameter is provided, '
                         'the currently active naming schemas are exported. You can provide this '
                         'parameter together with the -i parameter: in this case, the curren state '
                         'is written to the output file before doing the import'),
        make_option('-i', '--input-file',
                    action='store',
                    dest='input_file',
                    default=None,
                    help='File from which naming info is to be read.'),
        make_option('-r', '--restrict-to-family',
                    action='store',
                    dest='restrict_regexp',
                    default=None,
                    help='Regular expression by which the InfoObject types for which' \
                         'import/export is to be performed is restricted: only those InfoObjects are touched ' \
                         'types whose family name contains a match of the provided regular expression.'),


    )


    def __init__(self, *args, **kwargs):
        self.schemas = kwargs.get('schemas', None)
        try:
            del (kwargs['schemas'])
        except:
            pass
        super(Command,self).__init__(*args,**kwargs)

    def handle(self, *args, **options):

        iobject_types = InfoObjectType.objects.all()


        if not 'input_list' in options:
            options['input_list'] = None

        if options.get('output_file',None) and \
                (options.get('input_file',None) == options['output_file']):
            raise CommandError("Please specify different input and output files!")

        if options.get('restrict_regexp',None):
            re_restrict_family = re.compile(options['restrict_regexp'])

        if options.get('output_file',None):
            print "Exporting naming schemas"

            # Write the existing naming schemas to the output file
            result_list = []
            for iobject_type in iobject_types:
                if not options.get('restrict_regexp',None) or re_restrict_family.search(iobject_type.iobject_family.name):
                    schema_list = []
                    name_schemas = InfoObjectNaming.objects.filter(iobject_type=iobject_type).order_by(
                        'position').values_list('format_string', flat=True)
                    result_list.append((iobject_type.name, iobject_type.iobject_family.name, iobject_type.namespace.uri,list(name_schemas)))

            with open(options['output_file'], 'w') as outfile:
                json.dump(result_list, outfile, indent=4)

        if options.get('input_file',None) or options.get('input_list',None):
            print "Importing naming schemas."
            # Read naming schemas and import them
            if options['input_list']:
                naming_list = self.schemas
            else:
                with open(options['input_file'], 'r') as content_file:
                    json_string = content_file.read()
                    naming_list = json.loads(json_string)


            for iobject_type_info in naming_list:
                name,family_name,iodt_ns_uri,naming_list = iobject_type_info
                if not options.get('restrict_regexp',None) or re_restrict_family.search(family_name):
                    iodt_ns, created = DataTypeNameSpace.objects.get_or_create(uri=iodt_ns_uri)
                    iodt_family, created = InfoObjectFamily.objects.get_or_create(name=family_name)
                    iobject_type, created = InfoObjectType.objects.get_or_create(name=name,
                                                                        iobject_family=iodt_family,
                                                                        namespace=iodt_ns)
                    # Delete existing naming schemas
                    InfoObjectNaming.objects.filter(iobject_type=iobject_type).delete()
                    counter = 0
                    # Create naming schemas as specified in input file
                    for naming_schema in naming_list:
                        InfoObjectNaming.objects.create(iobject_type=iobject_type,
                                                        position=counter,
                                                        format_string=naming_schema)

                        counter += 1






# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


from django import template
from django.utils.html import strip_tags
from django.core.paginator import Paginator, PageNotAnInteger, EmptyPage

from dingos import DINGOS_TEMPLATE_FAMILY



register = template.Library()


def node_indent(elt_name, node_id, fact_term, attribute, highlight_node=None):
    """
    This tag uses a table structure to display indentation
    of fact terms based on the information contained in the
    node identifier.

    This tag and the closing 'node_indent_end' tag must
    enclose the value to be displayed after the display
    of the fact term.

    """
    # Some colors to chose from:
    color_dict = {0: {0: '#004C80', # blueish
                      1: '#005C99',
                      2: '#006BB2',
                      3: '#007ACC',
                      4: '#008AE6',
                      5: '#0099FF',
                      6: '#19A3FF',
                      7: '#33ADFF',
                      8: '#4DB8FF',
                      9: '#66C2FF',
                      10: '#80CCFF',
                      11: '#99D6FF',
                      12: '#B2E0FF',
                      13: '#CCEBFF',
                      14: '#E6F5FF'},
                  2: {0: '#008000', # greenish
                      1: '#009900',
                      2: '#00B200',
                      3: '#00CC00',
                      4: '#00E600',
                      5: '#00FF00',
                      6: '#19FF19',
                      7: '#33FF33',
                      8: '#4DFF4D',
                      9: '#66FF66',
                      10: '#80FF80',
                      11: '#99FF99',
                      12: '#B2FFB2',
                      13: '#CCFFCC',
                      14: '#E6FFE6'},
                  3: {0: '#804C80', # pinkish
                      1: '#995C99',
                      2: '#B26BB2',
                      3: '#CC7ACC',
                      4: '#E68AE6',
                      5: '#FF99FF',
                      6: '#FFA3FF',
                      7: '#FFADFF',
                      8: '#FFB8FF',
                      9: '#FFC2FF',
                      10: '#FFCCFF',
                      11: '#FFD6FF',
                      12: '#FFE0FF',
                      13: '#FFEBFF',
                      14: '#FFF5FF', },
                  1: {0: "#5C3D99", # violetish
                      1: "#6B47B2",
                      2: "#7A52CC",
                      3: "#8A5CE6",
                      4: "#9966FF",
                      5: "#A375FF",
                      6: "#AD85FF",
                      7: "#B894FF",
                      8: "#C2A3FF",
                      9: "#CCB2FF",
                      10: "#D6C2FF",
                      11: "#E0D1FF",
                      12: "#EBE0FF",
                      13: "#F5F0FF",
                      14: "#FFFFFF"}

    }

    indents = 100
    node_ids = node_id.split(':')
    fact_components = fact_term.split('/')

    if len(fact_components) == 1 and fact_components[0] == '':
        fact_components = []
    if attribute:
        fact_components.append("@%s" % attribute)

    fact_components = dict([(x, fact_components[x]) for x in range(0, len(fact_components))])


    #node_ids.reverse()

    result = []
    counter = 0
    for node in node_ids:
        is_attr = False
        if len(node) >= 1:
            if node[0] == 'A':
                is_attr = True

            node = node[1:]

        if len(node) > 0:
            node_nr = int(node)
        else:
            node_nr = 0
        if is_attr:
            node_mod = 2
        else:
            node_mod = node_nr % 2
        if is_attr:
            result.append("<%(elt_name)s style='background: %(color)s'>%(fact_term_component)s</%(elt_name)s>" % {
                'elt_name': elt_name,
                'fact_term_component': fact_components.get(counter, ''),
                'color': color_dict[2][max(14 - counter,4)]})
        else:
            result.append(
                "<%(elt_name)s style='width:1px; margin: 0px ; background : %(color)s'>%(fact_term_component)s</%(elt_name)s>" % {
                    'elt_name': elt_name,
                    'color': color_dict[node_mod][max(14 - counter,4)],
                    'fact_term_component': fact_components.get(counter, '')})

        counter += 1

    highlight = "style='background: #FF0000;'" if highlight_node == node_id else None

    result.append("<%(elt_name)s colspan='%(colspan)s' %(highlight)s>" % {'elt_name': elt_name, 'colspan': (indents - counter), 'highlight' : highlight})

    return "".join(result)


register.simple_tag(node_indent)


def node_indent_end(elt_name, node_id, fact_term, attribute):
    """
    Closing tag for the node_indent tag. Currently, only the
    parameter elt_name is used: we keep the other around in
    case we later find out that we need to do some more
    closing stuff based on the contents of the other parameters
    also used in the opening tag.
    """
    return "</%s>" % elt_name

register.simple_tag(node_indent_end)


@register.filter
def sliceupto(value, upto):
    """
    An additional slice feature working with int variables
    which can be set within the context of a view.
    Usage in template: {% for obj in list|sliceupto:z %}
    """
    try:
        return value[0:upto]
    except (ValueError, TypeError):
        return value

@register.simple_tag
def create_title(*args):
    """
    Combines all given arguments to create clean title-tags values.
    All arguments are divided by a " " seperator and HTML tags
    are to be removed.
    """
    seperator = " "
    return strip_tags(seperator.join(args))


@register.inclusion_tag('dingos/%s/includes/_Paginator.html' % DINGOS_TEMPLATE_FAMILY)
def render_paginator(view,paginator,page_obj):
    request_string = view.get_query_string(remove=['page'])
    return {'request_string':request_string,'paginator':paginator,'page_obj':page_obj}

# Below we register template tags that display
# certain aspects of an InformationObject.

@register.inclusion_tag('dingos/%s/includes/_InfoObjectFactsDisplay.html'% DINGOS_TEMPLATE_FAMILY)
def show_InfoObject(iobject, iobject2facts, view=None, highlight=None,show_NodeID=False):
    page = view.request.GET.get('page')


    #iobject2facts_paginator = Paginator(iobject.fact_thru.all(),100)
    iobject2facts_paginator = Paginator(iobject2facts,200)
    if iobject2facts_paginator.num_pages == 1:
        is_paginated = False
    else:
        is_paginated = True
    try:
        iobject2facts = iobject2facts_paginator.page(page)
        is_paginated=True
    except PageNotAnInteger:
        # If page is not an integer, deliver first page.
        iobject2facts = iobject2facts_paginator.page(1)

    except EmptyPage:
        # If page is out of range (e.g. 9999), deliver last page of results.
        iobject2facts = iobject2facts_paginator.page(iobject2facts_paginator.num_pages)


    return {'object': iobject,
            'view' : view,
            'is_paginated' : is_paginated,
            'paginator' : iobject2facts_paginator,
            'page_obj' :iobject2facts,
            'highlight' : highlight,
            'show_NodeID' : show_NodeID,
            'iobject2facts_paginator':iobject2facts_paginator,
            'iobject2facts': iobject2facts}


@register.inclusion_tag('dingos/%s/includes/_InfoObjectRevisionListDisplay.html'% DINGOS_TEMPLATE_FAMILY)
def show_InfoObjectRevisions(iobject):
    return {'object': iobject}


@register.inclusion_tag('dingos/%s/includes/_InfoObjectRevisionListDisplay_vertical.html'% DINGOS_TEMPLATE_FAMILY)
def show_InfoObjectRevisions_vertical(iobject):
    return {'object': iobject}


@register.inclusion_tag('dingos/%s/includes/_InfoObjectEmbeddingDisplay.html'% DINGOS_TEMPLATE_FAMILY)
def show_InfoObjectEmbeddings(iobject):
    return {'object': iobject}


@register.inclusion_tag('dingos/%s/includes/_InfoObjectEmbeddingDisplay_vertical.html'% DINGOS_TEMPLATE_FAMILY)
def show_InfoObjectEmbeddings_vertical(iobject,max_embedded):
    return {'object': iobject, 'max_embedded' : max_embedded}


@register.inclusion_tag('dingos/%s/includes/_InfoObjectIDDataDisplay.html'% DINGOS_TEMPLATE_FAMILY)
def show_InfoObjectIDData(iobject, show_hyperlink=False,show_title=False):
    return {'object': iobject,
            'show_hyperlink': show_hyperlink,
            'show_title': show_title}


@register.inclusion_tag('dingos/%s/includes/_InfoObjectMarkingsListDisplay.html'% DINGOS_TEMPLATE_FAMILY)
def show_InfoObjectMarkings(iobject):
    return {'object': iobject}



# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from django.conf import settings
from django.core.files.storage import FileSystemStorage

import dingos

if settings.configured and 'DINGOS' in dir(settings):
    dingos.DINGOS_TEMPLATE_FAMILY = settings.DINGOS.get('TEMPLATE_FAMILY', dingos.DINGOS_TEMPLATE_FAMILY)

if settings.configured and 'DINGOS' in dir(settings):
    dingos.DINGOS_DEFAULT_ID_NAMESPACE_URI = settings.DINGOS.get('OWN_ORGANIZATION_ID_NAMESPACE',
                                                                 dingos.DINGOS_DEFAULT_ID_NAMESPACE_URI)

if settings.configured and 'DINGOS' in dir(settings):
    dingos.DINGOS_BLOB_ROOT = settings.DINGOS.get('BLOB_ROOT',None)

if not dingos.DINGOS_BLOB_ROOT:

    dingos.DINGOS_BLOB_STORAGE=None
    #raise NotImplementedError("Please configure a BLOB_ROOT  directory in the DINGOS settings (look "
    #                          "at how the MEDIA directory is defined and define an appropriate directory "
    #                          "for storing stuff that does not got into the database (usually very large "
    #                          "values) on the filesystem.")
else:

    dingos.DINGOS_BLOB_STORAGE = FileSystemStorage(location=dingos.DINGOS_BLOB_ROOT)
    # We do not want the blobs to be directly available via URL.
    # Reading the code it seems that setting 'base_url=None' in
    # the __init__ arguments does not help, because __init__
    # then choses the media URL as default url. So we have
    # to set it explicitly after __init__ is done.
    dingos.DINGOS_BLOB_STORAGE.base_url=None

if settings.configured and 'DINGOS' in dir(settings):
    dingos.DINGOS_MAX_VALUE_SIZE_WRITTEN_TO_VALUE_TABLE = settings.DINGOS.get('DINGOS_MAX_VALUE_SIZE_WRITTEN_TO_VALUE_TABLE',
                                                                             dingos.DINGOS_MAX_VALUE_SIZE_WRITTEN_TO_VALUE_TABLE)

if settings.configured and 'DINGOS' in dir(settings):
    configured_large_value_dest = settings.DINGOS.get('LARGE_VALUE_DESTINATION',None)
    if configured_large_value_dest:
        if configured_large_value_dest == 'DINGOS_VALUES_TABLE':
            dingos.DINGOS_LARGE_VALUE_DESTINATION = dingos.DINGOS_VALUES_TABLE
    elif configured_large_value_dest == 'DINGOS_FILE_SYSTEM':
        dingos.DINGOS_LARGE_VALUE_DESTINATION = dingos.DINGOS_FILE_SYSTEM
    elif configured_large_value_dest == 'DINGOS_BLOB_TABLE':
        dingos.DINGOS_LARGE_VALUE_DESTINATION = dingos.DINGOS_BLOB_TABLE





# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from urlparse import urlparse

def extract_fqdn(uri):
    """ Extract the FQDN from an URI."""
    try:
        parsed = urlparse(uri)
        fqdn = parsed.netloc
        return fqdn
    except:
        return None



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'FactValue'
        db.create_table(u'dingos_factvalue', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('value', self.gf('django.db.models.fields.TextField')()),
            ('title', self.gf('django.db.models.fields.CharField')(max_length=256, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(blank=True)),
            ('fact_data_type', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['dingos.FactDataType'])),
        ))
        db.send_create_signal(u'dingos', ['FactValue'])

        # Adding unique constraint on 'FactValue', fields ['value', 'fact_data_type']
        db.create_unique(u'dingos_factvalue', ['value', 'fact_data_type_id'])

        # Adding model 'FactDataType'
        db.create_table(u'dingos_factdatatype', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=128)),
            ('description', self.gf('django.db.models.fields.TextField')(blank=True)),
            ('name_space', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['dingos.DataTypeNameSpace'])),
            ('kind', self.gf('django.db.models.fields.SmallIntegerField')(default=0)),
        ))
        db.send_create_signal(u'dingos', ['FactDataType'])

        # Adding unique constraint on 'FactDataType', fields ['name', 'name_space']
        db.create_unique(u'dingos_factdatatype', ['name', 'name_space_id'])

        # Adding model 'DataTypeNameSpace'
        db.create_table(u'dingos_datatypenamespace', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('uri', self.gf('django.db.models.fields.URLField')(unique=True, max_length=255)),
            ('name', self.gf('django.db.models.fields.SlugField')(max_length=50, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(blank=True)),
        ))
        db.send_create_signal(u'dingos', ['DataTypeNameSpace'])

        # Adding model 'IdentifierNameSpace'
        db.create_table(u'dingos_identifiernamespace', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('uri', self.gf('django.db.models.fields.URLField')(unique=True, max_length=255)),
            ('name', self.gf('django.db.models.fields.SlugField')(max_length=50, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(blank=True)),
        ))
        db.send_create_signal(u'dingos', ['IdentifierNameSpace'])

        # Adding model 'FactTerm'
        db.create_table(u'dingos_factterm', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('term', self.gf('django.db.models.fields.CharField')(max_length=512)),
            ('attribute', self.gf('django.db.models.fields.CharField')(max_length=128)),
        ))
        db.send_create_signal(u'dingos', ['FactTerm'])

        # Adding unique constraint on 'FactTerm', fields ['term', 'attribute']
        db.create_unique(u'dingos_factterm', ['term', 'attribute'])

        # Adding model 'InfoObjectNaming'
        db.create_table(u'dingos_infoobjectnaming', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('iobject_type', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['dingos.InfoObjectType'])),
            ('format_string', self.gf('django.db.models.fields.TextField')()),
            ('position', self.gf('django.db.models.fields.PositiveSmallIntegerField')()),
        ))
        db.send_create_signal(u'dingos', ['InfoObjectNaming'])

        # Adding model 'InfoObjectType'
        db.create_table(u'dingos_infoobjecttype', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.SlugField')(max_length=30)),
            ('description', self.gf('django.db.models.fields.TextField')(blank=True)),
            ('iobject_family', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['dingos.InfoObjectFamily'])),
            ('namespace', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['dingos.DataTypeNameSpace'], blank=True)),
        ))
        db.send_create_signal(u'dingos', ['InfoObjectType'])

        # Adding unique constraint on 'InfoObjectType', fields ['name', 'iobject_family', 'namespace']
        db.create_unique(u'dingos_infoobjecttype', ['name', 'iobject_family_id', 'namespace_id'])

        # Adding model 'InfoObjectFamily'
        db.create_table(u'dingos_infoobjectfamily', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.SlugField')(unique=True, max_length=256)),
            ('title', self.gf('django.db.models.fields.CharField')(max_length=1024, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(blank=True)),
        ))
        db.send_create_signal(u'dingos', ['InfoObjectFamily'])

        # Adding model 'Revision'
        db.create_table(u'dingos_revision', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=32, blank=True)),
        ))
        db.send_create_signal(u'dingos', ['Revision'])

        # Adding model 'FactTerm2Type'
        db.create_table(u'dingos_factterm2type', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('fact_term', self.gf('django.db.models.fields.related.ForeignKey')(related_name='iobject_type_thru', to=orm['dingos.FactTerm'])),
            ('iobject_type', self.gf('django.db.models.fields.related.ForeignKey')(related_name='fact_term_thru', to=orm['dingos.InfoObjectType'])),
            ('title', self.gf('django.db.models.fields.CharField')(max_length=1024, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(blank=True)),
        ))
        db.send_create_signal(u'dingos', ['FactTerm2Type'])

        # Adding unique constraint on 'FactTerm2Type', fields ['iobject_type', 'fact_term']
        db.create_unique(u'dingos_factterm2type', ['iobject_type_id', 'fact_term_id'])

        # Adding M2M table for field fact_data_types on 'FactTerm2Type'
        db.create_table(u'dingos_factterm2type_fact_data_types', (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('factterm2type', models.ForeignKey(orm[u'dingos.factterm2type'], null=False)),
            ('factdatatype', models.ForeignKey(orm[u'dingos.factdatatype'], null=False))
        ))
        db.create_unique(u'dingos_factterm2type_fact_data_types', ['factterm2type_id', 'factdatatype_id'])

        # Adding model 'NodeID'
        db.create_table(u'dingos_nodeid', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=255)),
        ))
        db.send_create_signal(u'dingos', ['NodeID'])

        # Adding model 'InfoObject2Fact'
        db.create_table(u'dingos_infoobject2fact', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('iobject', self.gf('django.db.models.fields.related.ForeignKey')(related_name='fact_thru', to=orm['dingos.InfoObject'])),
            ('fact', self.gf('django.db.models.fields.related.ForeignKey')(related_name='iobject_thru', to=orm['dingos.Fact'])),
            ('attributed_fact', self.gf('django.db.models.fields.related.ForeignKey')(related_name='attributes', null=True, to=orm['dingos.InfoObject2Fact'])),
            ('node_id', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['dingos.NodeID'])),
        ))
        db.send_create_signal(u'dingos', ['InfoObject2Fact'])

        # Adding model 'Fact'
        db.create_table(u'dingos_fact', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('fact_term', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['dingos.FactTerm'])),
            ('value_iobject_id', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='value_of_set', null=True, to=orm['dingos.Identifier'])),
            ('value_iobject_ts', self.gf('django.db.models.fields.DateTimeField')(null=True)),
            ('value_on_disk', self.gf('django.db.models.fields.BooleanField')(default=False)),
        ))
        db.send_create_signal(u'dingos', ['Fact'])

        # Adding M2M table for field fact_values on 'Fact'
        db.create_table(u'dingos_fact_fact_values', (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('fact', models.ForeignKey(orm[u'dingos.fact'], null=False)),
            ('factvalue', models.ForeignKey(orm[u'dingos.factvalue'], null=False))
        ))
        db.create_unique(u'dingos_fact_fact_values', ['fact_id', 'factvalue_id'])

        # Adding model 'InfoObject'
        db.create_table(u'dingos_infoobject', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('identifier', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['dingos.Identifier'])),
            ('timestamp', self.gf('django.db.models.fields.DateTimeField')()),
            ('create_timestamp', self.gf('django.db.models.fields.DateTimeField')()),
            ('iobject_type', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['dingos.InfoObjectType'])),
            ('iobject_type_revision', self.gf('django.db.models.fields.related.ForeignKey')(related_name='+', to=orm['dingos.Revision'])),
            ('iobject_family', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['dingos.InfoObjectFamily'])),
            ('iobject_family_revision', self.gf('django.db.models.fields.related.ForeignKey')(related_name='+', to=orm['dingos.Revision'])),
            ('uri', self.gf('django.db.models.fields.URLField')(max_length=200, blank=True)),
            ('name', self.gf('django.db.models.fields.CharField')(default='Unnamed', max_length=255, blank=True)),
        ))
        db.send_create_signal(u'dingos', ['InfoObject'])

        # Adding unique constraint on 'InfoObject', fields ['identifier', 'timestamp']
        db.create_unique(u'dingos_infoobject', ['identifier_id', 'timestamp'])

        # Adding model 'Identifier'
        db.create_table(u'dingos_identifier', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('uid', self.gf('django.db.models.fields.SlugField')(max_length=255)),
            ('namespace', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['dingos.IdentifierNameSpace'])),
            ('latest', self.gf('django.db.models.fields.related.OneToOneField')(related_name='latest_of', unique=True, null=True, to=orm['dingos.InfoObject'])),
        ))
        db.send_create_signal(u'dingos', ['Identifier'])

        # Adding unique constraint on 'Identifier', fields ['uid', 'namespace']
        db.create_unique(u'dingos_identifier', ['uid', 'namespace_id'])

        # Adding model 'Relation'
        db.create_table(u'dingos_relation', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('source_id', self.gf('django.db.models.fields.related.ForeignKey')(related_name='yields_via', null=True, to=orm['dingos.Identifier'])),
            ('target_id', self.gf('django.db.models.fields.related.ForeignKey')(related_name='yielded_by_via', null=True, to=orm['dingos.Identifier'])),
            ('relation_type', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['dingos.Fact'])),
            ('metadata_id', self.gf('django.db.models.fields.related.ForeignKey')(related_name='+', null=True, to=orm['dingos.Identifier'])),
        ))
        db.send_create_signal(u'dingos', ['Relation'])

        # Adding unique constraint on 'Relation', fields ['source_id', 'target_id', 'relation_type']
        db.create_unique(u'dingos_relation', ['source_id_id', 'target_id_id', 'relation_type_id'])

        # Adding model 'Marking2X'
        db.create_table(u'dingos_marking2x', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('marking', self.gf('django.db.models.fields.related.ForeignKey')(related_name='marked_item_thru', to=orm['dingos.InfoObject'])),
            ('content_type', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['contenttypes.ContentType'])),
            ('object_id', self.gf('django.db.models.fields.PositiveIntegerField')()),
        ))
        db.send_create_signal(u'dingos', ['Marking2X'])


    def backwards(self, orm):
        # Removing unique constraint on 'Relation', fields ['source_id', 'target_id', 'relation_type']
        db.delete_unique(u'dingos_relation', ['source_id_id', 'target_id_id', 'relation_type_id'])

        # Removing unique constraint on 'Identifier', fields ['uid', 'namespace']
        db.delete_unique(u'dingos_identifier', ['uid', 'namespace_id'])

        # Removing unique constraint on 'InfoObject', fields ['identifier', 'timestamp']
        db.delete_unique(u'dingos_infoobject', ['identifier_id', 'timestamp'])

        # Removing unique constraint on 'FactTerm2Type', fields ['iobject_type', 'fact_term']
        db.delete_unique(u'dingos_factterm2type', ['iobject_type_id', 'fact_term_id'])

        # Removing unique constraint on 'InfoObjectType', fields ['name', 'iobject_family', 'namespace']
        db.delete_unique(u'dingos_infoobjecttype', ['name', 'iobject_family_id', 'namespace_id'])

        # Removing unique constraint on 'FactTerm', fields ['term', 'attribute']
        db.delete_unique(u'dingos_factterm', ['term', 'attribute'])

        # Removing unique constraint on 'FactDataType', fields ['name', 'name_space']
        db.delete_unique(u'dingos_factdatatype', ['name', 'name_space_id'])

        # Removing unique constraint on 'FactValue', fields ['value', 'fact_data_type']
        db.delete_unique(u'dingos_factvalue', ['value', 'fact_data_type_id'])

        # Deleting model 'FactValue'
        db.delete_table(u'dingos_factvalue')

        # Deleting model 'FactDataType'
        db.delete_table(u'dingos_factdatatype')

        # Deleting model 'DataTypeNameSpace'
        db.delete_table(u'dingos_datatypenamespace')

        # Deleting model 'IdentifierNameSpace'
        db.delete_table(u'dingos_identifiernamespace')

        # Deleting model 'FactTerm'
        db.delete_table(u'dingos_factterm')

        # Deleting model 'InfoObjectNaming'
        db.delete_table(u'dingos_infoobjectnaming')

        # Deleting model 'InfoObjectType'
        db.delete_table(u'dingos_infoobjecttype')

        # Deleting model 'InfoObjectFamily'
        db.delete_table(u'dingos_infoobjectfamily')

        # Deleting model 'Revision'
        db.delete_table(u'dingos_revision')

        # Deleting model 'FactTerm2Type'
        db.delete_table(u'dingos_factterm2type')

        # Removing M2M table for field fact_data_types on 'FactTerm2Type'
        db.delete_table('dingos_factterm2type_fact_data_types')

        # Deleting model 'NodeID'
        db.delete_table(u'dingos_nodeid')

        # Deleting model 'InfoObject2Fact'
        db.delete_table(u'dingos_infoobject2fact')

        # Deleting model 'Fact'
        db.delete_table(u'dingos_fact')

        # Removing M2M table for field fact_values on 'Fact'
        db.delete_table('dingos_fact_fact_values')

        # Deleting model 'InfoObject'
        db.delete_table(u'dingos_infoobject')

        # Deleting model 'Identifier'
        db.delete_table(u'dingos_identifier')

        # Deleting model 'Relation'
        db.delete_table(u'dingos_relation')

        # Deleting model 'Marking2X'
        db.delete_table(u'dingos_marking2x')


    models = {
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'dingos.datatypenamespace': {
            'Meta': {'object_name': 'DataTypeNameSpace'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.SlugField', [], {'max_length': '50', 'blank': 'True'}),
            'uri': ('django.db.models.fields.URLField', [], {'unique': 'True', 'max_length': '255'})
        },
        u'dingos.fact': {
            'Meta': {'object_name': 'Fact'},
            'fact_term': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.FactTerm']"}),
            'fact_values': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['dingos.FactValue']", 'null': 'True', 'symmetrical': 'False'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'value_iobject_id': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'value_of_set'", 'null': 'True', 'to': u"orm['dingos.Identifier']"}),
            'value_iobject_ts': ('django.db.models.fields.DateTimeField', [], {'null': 'True'}),
            'value_on_disk': ('django.db.models.fields.BooleanField', [], {'default': 'False'})
        },
        u'dingos.factdatatype': {
            'Meta': {'unique_together': "(('name', 'name_space'),)", 'object_name': 'FactDataType'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'kind': ('django.db.models.fields.SmallIntegerField', [], {'default': '0'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'name_space': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.DataTypeNameSpace']"})
        },
        u'dingos.factterm': {
            'Meta': {'unique_together': "(('term', 'attribute'),)", 'object_name': 'FactTerm'},
            'attribute': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'term': ('django.db.models.fields.CharField', [], {'max_length': '512'})
        },
        u'dingos.factterm2type': {
            'Meta': {'unique_together': "(('iobject_type', 'fact_term'),)", 'object_name': 'FactTerm2Type'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'fact_data_types': ('django.db.models.fields.related.ManyToManyField', [], {'related_name': "'fact_term_thru'", 'symmetrical': 'False', 'to': u"orm['dingos.FactDataType']"}),
            'fact_term': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'iobject_type_thru'", 'to': u"orm['dingos.FactTerm']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'iobject_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'fact_term_thru'", 'to': u"orm['dingos.InfoObjectType']"}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '1024', 'blank': 'True'})
        },
        u'dingos.factvalue': {
            'Meta': {'unique_together': "(('value', 'fact_data_type'),)", 'object_name': 'FactValue'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'fact_data_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.FactDataType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '256', 'blank': 'True'}),
            'value': ('django.db.models.fields.TextField', [], {})
        },
        u'dingos.identifier': {
            'Meta': {'unique_together': "(('uid', 'namespace'),)", 'object_name': 'Identifier'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'latest': ('django.db.models.fields.related.OneToOneField', [], {'related_name': "'latest_of'", 'unique': 'True', 'null': 'True', 'to': u"orm['dingos.InfoObject']"}),
            'namespace': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.IdentifierNameSpace']"}),
            'uid': ('django.db.models.fields.SlugField', [], {'max_length': '255'})
        },
        u'dingos.identifiernamespace': {
            'Meta': {'object_name': 'IdentifierNameSpace'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.SlugField', [], {'max_length': '50', 'blank': 'True'}),
            'uri': ('django.db.models.fields.URLField', [], {'unique': 'True', 'max_length': '255'})
        },
        u'dingos.infoobject': {
            'Meta': {'ordering': "['-timestamp']", 'unique_together': "(('identifier', 'timestamp'),)", 'object_name': 'InfoObject'},
            'create_timestamp': ('django.db.models.fields.DateTimeField', [], {}),
            'facts': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['dingos.Fact']", 'through': u"orm['dingos.InfoObject2Fact']", 'symmetrical': 'False'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'identifier': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.Identifier']"}),
            'iobject_family': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.InfoObjectFamily']"}),
            'iobject_family_revision': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['dingos.Revision']"}),
            'iobject_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.InfoObjectType']"}),
            'iobject_type_revision': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['dingos.Revision']"}),
            'name': ('django.db.models.fields.CharField', [], {'default': "'Unnamed'", 'max_length': '255', 'blank': 'True'}),
            'timestamp': ('django.db.models.fields.DateTimeField', [], {}),
            'uri': ('django.db.models.fields.URLField', [], {'max_length': '200', 'blank': 'True'})
        },
        u'dingos.infoobject2fact': {
            'Meta': {'ordering': "['node_id__name']", 'object_name': 'InfoObject2Fact'},
            'attributed_fact': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'attributes'", 'null': 'True', 'to': u"orm['dingos.InfoObject2Fact']"}),
            'fact': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'iobject_thru'", 'to': u"orm['dingos.Fact']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'iobject': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'fact_thru'", 'to': u"orm['dingos.InfoObject']"}),
            'node_id': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.NodeID']"})
        },
        u'dingos.infoobjectfamily': {
            'Meta': {'object_name': 'InfoObjectFamily'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.SlugField', [], {'unique': 'True', 'max_length': '256'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '1024', 'blank': 'True'})
        },
        u'dingos.infoobjectnaming': {
            'Meta': {'ordering': "['position']", 'object_name': 'InfoObjectNaming'},
            'format_string': ('django.db.models.fields.TextField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'iobject_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.InfoObjectType']"}),
            'position': ('django.db.models.fields.PositiveSmallIntegerField', [], {})
        },
        u'dingos.infoobjecttype': {
            'Meta': {'unique_together': "(('name', 'iobject_family', 'namespace'),)", 'object_name': 'InfoObjectType'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'iobject_family': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.InfoObjectFamily']"}),
            'name': ('django.db.models.fields.SlugField', [], {'max_length': '30'}),
            'namespace': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.DataTypeNameSpace']", 'blank': 'True'})
        },
        u'dingos.marking2x': {
            'Meta': {'object_name': 'Marking2X'},
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'marking': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'marked_item_thru'", 'to': u"orm['dingos.InfoObject']"}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {})
        },
        u'dingos.nodeid': {
            'Meta': {'object_name': 'NodeID'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255'})
        },
        u'dingos.relation': {
            'Meta': {'unique_together': "(('source_id', 'target_id', 'relation_type'),)", 'object_name': 'Relation'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'metadata_id': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'null': 'True', 'to': u"orm['dingos.Identifier']"}),
            'relation_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.Fact']"}),
            'source_id': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'yields_via'", 'null': 'True', 'to': u"orm['dingos.Identifier']"}),
            'target_id': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'yielded_by_via'", 'null': 'True', 'to': u"orm['dingos.Identifier']"})
        },
        u'dingos.revision': {
            'Meta': {'object_name': 'Revision'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '32', 'blank': 'True'})
        }
    }

    complete_apps = ['dingos']


# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#



import logging
import re

from django.utils.datastructures import SortedDict

logger = logging.getLogger(__name__)



def dict2tuple(mydict):
    """
    Turn a dictionary of the form used by DINGO
    (i.e., a key is mapped to either another dictionary,
    a list or a value) into a tuple-representation.
    This is useful for providing a standard python object that
    observes the order of elements (which standard dictionaries don't).
    """
    result = []
    for key in mydict.keys():
        # There are three cases we need to consider:
        # dictionary, list, and other values
        try:
            # We test whether the object is a dictionary-like
            # structure via accessing the
            keys = mydict[key].keys()
            key_result = dict2tuple(mydict[key])
        except AttributeError:
            # So we have either a list or a value
            if type(mydict[key]) == type([]):
                key_result = []
                for elt in mydict[key]:
                    key_result.append(dict2tuple(elt))
            else:
                key_result = mydict[key]
        result.append((key, key_result))
    return tuple(result)


def tuple2dict(mytuple, constructor=dict):
    """
    Turn a tuple as returned by dict2tuple back
    into a dictionary structure. The optional
    parameter "constructor" governs, what class
    is used to create the dictionary: by default,
    the standard dict class is used, but in DINGO
    we mostly use the DingoObjDict class.
    """
    result = constructor()
    for elt in mytuple:
        key, value = elt
        if type(value) == type(()):
            key_value = tuple2dict(value, constructor=constructor)
        elif type(value) == type([]):
            key_value = []
            for elt in value:
                key_value.append(tuple2dict(elt, constructor=constructor))
        else:
            key_value = value
        result[key] = key_value
    return result


class ExtendedSortedDict(SortedDict):
    """
    ExtendedSortedDict adds a few convenient
    methods to Django's SortedDict class.
    """
    #def copy(self):
    #    obj = self.__class__(self)
    #    obj.keyOrder = self.keyOrder[:]
    #    return obj

    def to_tuple(self):
        """
        Turn an extended sorted dictionary of the form used by DINGO
        (i.e., a key is mapped to either another dictionary,
        a list or a value) into a tuple-representation.
        This is useful for providing a standard python object that
        observes the order of elements (which standard dictionaries don't).

        """
        return dict2tuple(self)

    def to_dict(self):
        """
        Turn an extended sorted dictionary of the form used by DINGO
        (i.e., a key is mapped to either another dictionary,
        a list or a value) into a normal dictionary (i.e., the
        key order is lost).
        """
        return tuple2dict(dict2tuple(self))


    def chained_get(self, *keys):
        """
        This function allows traversals over several keys to
        be performed by passing a list of keys::

            d.chained_get(key1,key2,key3) = d[key1][key2][key3]

        """
        existing = self
        for i in range(0, len(keys)):
            if keys[i] in existing:
                existing = existing[keys[i]]
            else:
                return None
        return existing

    def chained_set(self, value, command='set', *keys):
        """
        chained_set takes the value to
        enter into the dictionary, a command of what
        to do with the value, and a sequence of keys.

        Examples:

        d = {}

        d.chained_set(1,'append','level 1','level 2')

        -> d['level 1']['level 2'] = [1]

        d.chained_set(2,'append','level 1','level 2')

        -> d['level 1']['level 2'] = [1,2]

        """
        new_object = self.__class__()
        existing = self
        for i in range(0, len(keys) - 1):
            if keys[i] in existing:
                existing = existing[keys[i]]
            else:
                existing[keys[i]] = new_object
                existing = existing[keys[i]]
        if command == 'set':
            existing[keys[len(keys) - 1]] = value
        elif command == 'append':
            if keys[len(keys) - 1] in existing:
                existing[keys[len(keys) - 1]].append(value)
            else:
                existing[keys[len(keys) - 1]] = [value]

        elif command == 'set_or_append':
            if keys[len(keys) - 1] in existing:
                if type(keys[len(keys) - 1]) == type([]):
                    existing[keys[len(keys) - 1]].append(value)
                else:
                    existing[keys[len(keys) - 1]] = [existing[keys[len(keys) - 1]], value]
            else:
                existing[keys[len(keys) - 1]] = value
        elif command == 'insert':
            if keys[len(keys) - 1] in existing:
                if not value in existing[keys[len(keys) - 1]]:
                    existing[keys[len(keys) - 1]].append(value)
            else:
                existing[keys[len(keys) - 1]] = [value]


class DingoObjDict(ExtendedSortedDict):
    """
    The DingoObjDict extends the ExtendedSortedDict with
    a flatten function that flattens out the dictionary
    into a list of facts. For details, see below in the
    documentation of the flatten function.
    """

    # TODO Remove commented-out code
    #def dictify(self):
    #   return dict([(k, (v.dictify() if isinstance(v,dict) else v))
    #                  for (k,v) in self.items()])

    #def __repr__(self):
    #
    #    old_dict = self.dictify()
    #    return old_dict.__repr__()



    def flatten(self, attr_ignore_predicate=None,force_nonleaf_fact_predicate=None):
        """
        Flatten a Dingo dictionary representation of into a list
        of fact-term/value pairs and associated information about tree structure (in node_id)
        and a dictionary mapping XML attributes to node identifiers.

        This all is best explained by example (see below).

        Note that attributes are also represented in the flattened representation. There
        are cases in which attributes should not lead to an entry in the flattened representation.
        This can be configured by passing the flatten method a predicate function 'attr_ignore_predicate'.
        The function takes a dictionary and returns True (ignore) or False (do not ignore). It
        is passed the dictionary representing an entry in the flat representation, such as follows::

              { 'node_id': 'N001:L000:N000:A000',
              'term': 'Hashes/Hash/Simple_Hash_Value',
              'attribute': 'condition',
              'value': u'Equals'},

        By default, the following predicate is used::

             (lambda x : '@' in x['attribute'])

        This is because during import, DINGO adds certain annotations to the import dictionary
        as attributes with a second leading '@'.

        Input Example (CybOx)::

            {'File_Name': {'@condition': 'Equals',  '_value': u'UNITED NATIONS COMPENSATION SCHEME...pdf'},
              'Hashes': {'Hash': [{'Simple_Hash_Value':
                                    {'@condition': 'Equals',
                                     '@datatype': 'hexBinary',
                                      '_value': u'576fea79dd23a352a14c3f8bf3dbc9eb732e1d54f804a29160894aec55df4bd5'},
                                   'Type': {'_value': 'SHA256'}},
                                  {'Simple_Hash_Value': {'@condition': 'Equals',
                                                         '@datatype': 'hexBinary',
                                                         '_value': u'491809c2092cecd633e43d465409a78c'},
                                   'Type': {'_value': 'MD5'}}]
                        }
            }

        Output Example::

          [ { 'node_id': 'N000',
              'term': 'File_Name',
              'attribute': False,
              'value': u'UNITED NATIONS COMPENSATION SCHEME...pdf'},
              { 'node_id': 'N000:A000',
              'term': 'File_Name',
              'attribute': 'condition',
              'value': u'Equals'},
            { 'node_id': 'N001:L000:N000',
              'term': 'Hashes/Hash/Simple_Hash_Value',
              'attribute': False,
              'value': u'576fea79dd23a352a14c3f8bf3dbc9eb732e1d54f804a29160894aec55df4bd5'},
            { 'node_id': 'N001:L000:N000:A000',
              'term': 'Hashes/Hash/Simple_Hash_Value',
              'attribute': 'condition',
              'value': u'Equals'},
            { 'node_id': 'N001:L000:N000:A001',
              'term': 'Hashes/Hash/Simple_Hash_Value',
              'attribute': 'datatype',
              'value': u'hexBinary'},
            { 'node_id': 'N001:L000:N001',
              'term': 'Hashes/Hash/Type',
              'attribute': False,
              'value': 'SHA256'},
            { 'node_id': 'N001:L001:N000',
              'term': 'Hashes/Hash/Simple_Hash_Value',
              'attribute': False,
              'value': u'491809c2092cecd633e43d465409a78c'},
            { 'node_id': 'N001:L001:N000:A000',
              'term': 'Hashes/Hash/Simple_Hash_Value',
              'attribute': 'condition',
              'value': u'Equals'},
            { 'node_id': 'N001:L001:N000:A001',
              'term': 'Hashes/Hash/Simple_Hash_Value',
              'attribute': 'datatype',
              'value': u'hexBinary'},
            { 'node_id': 'N001:L001:N001',
              'term': 'Hashes/Hash/Type',
              'attribute': False,
              'value': 'MD5'}
          ]

          { 'N000': { 'condition': 'Equals'},
            'N001:L000:N000': { 'condition': 'Equals', 'datatype': 'hexBinary'},
            'N001:L001:N000': { 'condition': 'Equals', 'datatype': 'hexBinary'},
          }

        """


        def node_id_gen(n):
            """
            Given an integer, generate a fixed-length representation.

            This is used for coding tree-node representations such
            as N001:N005:N009 (9th child of 5th child of 1st child).

            Currently, we use three digits, i.e., we get problems
            if we import stuff with more than 1000 children in a node.

            """

            return "%s%04d" % (n[0], n[1])



        RE_ELEMENT_MATCHER = re.compile(r"[^@_].*")


        def _flatten(self, result_list, attr_dict, namespace, prefix):
            """
            Flatten a Dingo dictionary representation of an infomration object into a list
            of fact-term/value pairs and associated information about tree structure (in node_id)
            and XML attributes.

            Internal function for recursive calls.

            """

            attributes = filter(lambda x: x[0] == '@', self.keys())

            node_id = ':'.join(map(node_id_gen, prefix))
            for attribute in attributes:
                if node_id not in attr_dict.keys():
                    attr_dict[node_id] = {attribute[1:]: self[attribute]}
                else:
                    attr_dict[node_id][attribute[1:]] = self[attribute]

            elements = filter(lambda x: RE_ELEMENT_MATCHER.match(x), self)

            fact_data = {'term': '/'.join(namespace),
                         'value': self.get('_value', ''),
                         'attribute': False,
                         'node_id': node_id}

            if elements == [] or force_nonleaf_fact_predicate(fact_data,attributes):
                logger.debug("Entered _VALUE branch for %s" % self)
                if '_value' in self.keys() or attributes != []:
                    fact_data = {'term': '/'.join(namespace),
                                 'value': self.get('_value', ''),
                                 'attribute': False,
                                 'node_id': node_id}
                    result_list.append(fact_data)
                    logger.debug("Appended fact %s" % fact_data)
            if elements != []:

                counter = 0
                for element in elements:
                    logger.debug("Processing element %s" % element)
                    if type(self[element]) == type([]):
                        logger.debug("Entered list branch for %s " % self[element])
                        for sub_elt in self[element]:
                            (result_list, attr_dict) = _flatten(sub_elt,
                                                                result_list=result_list,
                                                                attr_dict=attr_dict,
                                                                namespace=namespace + [element],
                                                                prefix=prefix + [('L', counter)])


                            counter += 1
                    elif type(self[element]) == type("hallo") or type(self[element]) == unicode:
                        logger.debug("Entered value branch for %s" % self[element])
                        # added this branch to deal with abbreviated dictionaries
                        # that provide value direcly rather then via '_value' key in dictionary

                        # temporarily append namespace
                        namespace.append(element)

                        fact_data = {'term': '/'.join(namespace),
                                     'value': self[element],
                                     'attribute': False,
                                     'node_id': "%s" % ':'.join(map(node_id_gen, prefix + [('N', counter)]))}
                        logger.debug("Appended fact %s" % fact_data)
                        result_list.append(fact_data)
                        # clean up namespace
                        namespace = namespace[:-1]
                        counter += 1
                    else:
                        logger.debug("Recursing for %s" % self[element])
                        (result_list, attr_dict) = _flatten(self[element],
                                                            result_list=result_list,
                                                            attr_dict=attr_dict,
                                                            namespace=namespace + [element],
                                                            prefix=prefix + [('N', counter)])
                        counter += 1

            attr_counter = 0
            for attribute in attributes:
                if node_id == '':
                    attr_node_id = node_id_gen(('A', attr_counter))
                else:
                    attr_node_id = "%s:%s" % (node_id, node_id_gen(('A', attr_counter)))
                fact = {'term': "%s" % ('/'.join(namespace)),
                        'value': self[attribute],
                        'node_id': attr_node_id,
                        'attribute': attribute[1:]}
                if not attr_ignore_predicate(fact):
                    result_list.append(fact)
                    logger.debug("Appended fact %s" % fact)
                    attr_counter += 1
                else:
                    logger.debug("Ignoring fact %s because of attr_ignore_list" % fact)
            result_list.sort(key=lambda x: x['node_id'])

            return (result_list, attr_dict)

        if not attr_ignore_predicate:
            attr_ignore_predicate = (lambda x: '@' in x['attribute'])

        if not force_nonleaf_fact_predicate:
            force_nonleaf_fact_predicate = (lambda x,y: False)

        return _flatten(self,result_list=[], attr_dict={}, namespace=[], prefix=[])

    def from_flat_repr(self,fact_list,include_node_id=False):
        """
        Convert a flat representation of information (consisting of a fact list and a dictionary
        mapping node ids to attributes into a dictionary representation information.

        Input Example (CybOx)::


          { '00': { 'condition': 'Equals', 'datatype': 'String'},
            '01:L00:00': { 'condition': 'Equals', 'datatype': 'hexBinary'},
            '01:L00:01': { 'datatype': 'String'},
            '01:L01:00': { 'condition': 'Equals', 'datatype': 'hexBinary'},
            '01:L01:01': { 'datatype': 'String'}
          }

        Output Example::

            {'File_Name': {'@condition': 'Equals', '@datatype': 'String', '_value': u'UNITED NATIONS COMPENSATION SCHEME...pdf'},
              'Hashes': {'Hash': [{'Simple_Hash_Value':
                                    {'@condition': 'Equals',
                                     '@datatype': 'hexBinary',
                                      '_value': u'576fea79dd23a352a14c3f8bf3dbc9eb732e1d54f804a29160894aec55df4bd5'},
                                   'Type': {'@datatype': 'String', '_value': 'SHA256'}},
                                  {'Simple_Hash_Value': {'@condition': 'Equals',
                                                         '@datatype': 'hexBinary',
                                                         '_value': u'491809c2092cecd633e43d465409a78c'},
                                   'Type': {'@datatype': 'String', '_value': 'MD5'}}]
                        }
            }



        """

        def node_id_unpack(n):
            """
            Given a node_id, unpack it into list-signifier
            (if existing) and position information
            """
            return (n[0],int(n[1:]))

        fact_list.sort(key=lambda x : x['node_id'])

        result = self
        for fact in fact_list:
            fact_path = fact['term'].split('/')
            del(fact['term'])
            if fact.get('attribute'):
                if fact_path == [""]:
                    fact_path = []
                fact_path.append('@%s' % fact['attribute'])

                del(fact['attribute'])
            node_path = fact['node_id'].split(':')
            del(fact['node_id'])
            walker = result
            #logger.debug(fact)
            for i in range(0,len(node_path)):

                #logger.debug("%s  %s" % (node_path,node_path[i]))
                (node_kind,counter) = node_id_unpack(node_path[i])
                element = fact_path[i]
                node_id = ':'.join(node_path[0:i+1])


                if not (element in walker.keys()):

                    if node_kind == 'L':
                        walker[element] = []
                        child_dict = DingoObjDict()
                        if include_node_id:
                            child_dict['@@node_id'] = node_id
                        walker[element].append(child_dict)
                        walker = walker[element][0]
                    elif node_kind == 'N':
                        child_dict = DingoObjDict()
                        if include_node_id:
                            child_dict['@@node_id'] = node_id
                        walker[element] = child_dict
                        walker = walker[element]

                else:

                    if node_kind == 'L':

                        if len(walker[element]) == counter:
                            next_item = DingoObjDict()
                            if include_node_id:
                                next_item['@@node_id'] = node_id
                            walker[element].append(next_item)
                            walker=walker[element][counter]# next_item
                        elif len(walker[element]) > counter:
                            walker = walker[element][counter]
                        else:
                            # TODO: introduce error logging
                            logger.error("ERROR: Structural problem for %s, counter %s and node_id %s" % (element,counter,node_id))
                            pass

                    else:
                        walker=walker[element]
            if node_kind == 'A':
                if len(fact['value_list']) == 1:
                    value = fact['value_list'][0]
                else:
                    value = fact['value_list']
                if include_node_id:
                    walker[element] = {'@@node_id' : node_id,
                                       '_value' : value}
                else:
                    walker[element] = value
            else:
                if len(fact['value_list']) == 1:
                    if fact['value_list'][0] != '':
                        walker['_value'] = fact['value_list'][0]
                else:
                    walker['_value'] = fact['value_list']
                if include_node_id:
                    walker['@@node_id'] = node_id

                del(fact['value_list'])

                for key in fact:
                    if fact[key]:
                        walker[key] = fact[key]





def dict2DingoObjDict(data):
    """
    Turn a dictionary of the form used by DINGO
    (i.e., a key is mapped to either another dictionary,
    a list or a value) into a DingoObjDict.
    """
    info_tuple = dict2tuple(data)
    info_dict = tuple2dict(info_tuple, constructor=DingoObjDict)
    return info_dict



# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from django.views.generic.base import ContextMixin
from django.views.generic import DetailView, ListView, TemplateView

from braces.views import LoginRequiredMixin, SelectRelatedMixin,PrefetchRelatedMixin
from core.http_helpers import get_query_string

from django_filters.views import FilterView

from dingos import DINGOS_TEMPLATE_FAMILY


class CommonContextMixin(ContextMixin):
    def get_context_data(self, **kwargs):
        context = super(CommonContextMixin, self).get_context_data(**kwargs)
        context['title'] = self.title if hasattr(self, 'title') else '[TITLE MISSING]'

        return context

class ViewMethodMixin(object):
    """
    We use this Mixin to enrich view with methods that are required
    by certain templates and template tags. In order to use
    these template tags from a given view, simply add this
    mixin to the parent classes of the view.
    """
    def get_query_string(self,*args,**kwargs):
        """
        Allows access to query string (with facilities to manipulate the string,
        e.g., removing or adding query attributes. We use this, for example,
        in the paginator, which needs to create a link with the current
        query and change around the 'page' part of the query string.
        """
        return get_query_string(self.request,*args,**kwargs)

class BasicListView(CommonContextMixin,ViewMethodMixin,LoginRequiredMixin,ListView):

    login_url = "/admin"

    template_name = 'dingos/%s/lists/base_lists_two_column.html' % DINGOS_TEMPLATE_FAMILY

    breadcrumbs = ()

    paginate_by = 20

class BasicFilterView(CommonContextMixin,ViewMethodMixin,LoginRequiredMixin,FilterView):

    login_url = "/admin"

    template_name = 'dingos/%s/lists/base_lists_two_column.html' % DINGOS_TEMPLATE_FAMILY

    breadcrumbs = ()

    paginate_by = 20

class BasicDetailView(CommonContextMixin,
                      ViewMethodMixin,
                      LoginRequiredMixin,
                      SelectRelatedMixin,
                      PrefetchRelatedMixin,
                      DetailView):

    login_url = "/admin"

    select_related = ()
    prefetch_related = ()

    breadcrumbs = (('Dingo',None),
                   ('View',None),
    )


class BasicTemplateView(CommonContextMixin,
                       ViewMethodMixin,
                       LoginRequiredMixin,
                       TemplateView):

    login_url = "/admin"


    breadcrumbs = (('Dingo',None),
                   ('View',None),
    )





# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#




__version__ = '0.1.0'

REVISION = __version__


# Below, default values used in DINGO are defined.


# Dingos allows for extensions with another template basis othan the
# Grappelli templates that are used by default. Such an extension
# would have to create a 'dingos/templates/<extension_name>' directory
# and reimplement the existing Dingos template structure in that
# directory.

DINGOS_TEMPLATE_FAMILY  = 'grappelli'

# The namespace uri should be used for qualifying identifiers if
# neither namespace nor namespace uri is explicitly provided for an identifier.

DINGOS_DEFAULT_ID_NAMESPACE_URI = 'enter.a.value.in.settings'


# The DINGOS_MISSING_ID_NAMESPACE_URI_PREFIX should be used s prefix for namespace
# information used in identifiers for which no namespace uri can be determined.

DINGOS_MISSING_ID_NAMESPACE_URI_PREFIX = 'https://github.com/siemens/django-dingos/wiki/namespaces_identifiers_missing'


# The DINGOS_NAMESPACE is used as default namespace for
# datatypes and information-object types created by DINGO.

DINGOS_NAMESPACE_URI = 'https://github.com/siemens/django-dingos/wiki/namespaces_dingos_types'
DINGOS_NAMESPACE_SLUG = 'DingosDefaultNameSpace'

DINGOS_DEFAULT_FACT_DATATYPE = 'String'

# The DINGOS_ID_NAMESPACE_URI is used as qualifier for identifiers
# of internal objects created by DINGO, such as information
# objects containing meta data of relations.

DINGOS_ID_NAMESPACE_URI = 'https://github.com/siemens/django-dingos/wiki/namespaces_dingos_identifiers'
DINGOS_ID_NAMESPACE_SLUG = 'DingosDefaultIdNameSpace'


# The DINGOS_GENERIC_FAMILY is used to provide a default family name for generic imports

DINGOS_GENERIC_FAMILY_NAME = 'generic'

# The DINGOS_IOBJECT_FAMILY_NAME is used as family name for all internally created objects
# such as PLACEHOLDERS. We Dingo's revision both as revision for families and object types
# of internally created information objects.

DINGOS_IOBJECT_FAMILY_NAME = 'DINGOS'

DINGOS_REVISION_NAME = REVISION

# Below, we define names used for internally created objects:

DINGOS_RELATION_TYPE_FACTTERM_NAME = '@@RelationType'

DINGOS_RELATION_METADATA_OBJECT_TYPE_NAME = 'RelationMetadata'

DINGOS_PLACEHOLDER_TYPE_NAME = 'PLACEHOLDER'

DINGOS_DEFAULT_IMPORT_MARKING_TYPE_NAME = "ImportInfo"


DINGOS_CONFIGURATION_TYPE_NAME = 'USER_CONFIG'


# Values larger than DINGOS_MAX_VALUE_SIZE_WRITTEN_TO_DB are
# not written to the data base but stored on the file system or
# written to a special blob table

DINGOS_MAX_VALUE_SIZE_WRITTEN_TO_VALUE_TABLE = 2048


# values for the LARGE_VALUE_DESTINATION are 'BLOB_TABLE' and 'FILE_SYSTEM'

DINGOS_VALUES_TABLE = 0
DINGOS_FILE_SYSTEM = 1
DINGOS_BLOB_TABLE = 2

DINGOS_LARGE_VALUE_DESTINATION = DINGOS_BLOB_TABLE



# The DINGOS_BLOB_ROOT absolutely has to be set in the DINGOS settings.
# If that is not the case, the attempt to read the value from the settings

DINGOS_BLOB_ROOT = None

# read_settings.py will instantiate the DINGOS_BLOB_STORAGE with a file storage handler
# that can also be used by importers to write file content to disk.

DINGOS_BLOB_STORAGE = None






# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import logging
import hashlib
import getpass
import os
import re
import json
import glob
import traceback
from optparse import make_option

from django.core.management.base import BaseCommand, CommandError
from django.utils import timezone

from dingos.core.decorators import print_arguments

from dingos.core.datastructures import dict2DingoObjDict
from dingos import *
from dingos.import_handling import DingoImportHandling


DingoImporter = DingoImportHandling()

logger = logging.getLogger(__name__)


class Generic_XML_Import:
    """
    This class provides the xml_import function for
    generic XML import into DINGO.

    This class can be used
    as template for defining specialized XML importers
    that have knowledge of the XML they are importing
    (e.g., are able to determine identifiers and timestamps,
    extract embedded objects, etc.). Have a look
    at the MANTIS importer for IODEF for an example
    of an importer with only moderate amounts of
    configuration.
    """
    def __init__(self, *args, **kwargs):

        # We keep track of toplevel attributes
        # and namespace info

        self.toplevel_attrs = {}
        self.namespace_dict = {}


    #
    # Now, we define functions for the hooks provided to us
    # by the DINGO xml-import. For the generic import
    # there is almost nothing to be done.



    def id_and_revision_extractor(self, xml_elt):
        """
        Function for generating a unique identifier for extracted embedded content;
        to be used for DINGO's xml-import hook 'embedded_id_gen'.

        For generic import, this function should be called exactly once, namely
        on the root element, but we ignore the result and rather derive
        a identifier from the hash value of the imported file.

        """

        return {'id': None,
                'timestamp': None,
        }


    def cybox_embedding_pred(self, parent, child, ns_mapping):
        """
        Predicate for recognizing inlined content in an XML; to
        be used for DINGO's xml-import hook 'embedded_predicate'.

        The generic importer does not recognize anything as embedded.
        """

        return False


    def ft_handler_list(self):
        """
        The fact-term handler list consists of a pairs of predicates
        and functions to be called on facts before they are created.
        These handlers can modify facts. For the generic import,
        nothing is done.
        """
        return []

    def datatype_extractor(self, iobject, fact, attr_info, namespace_mapping, add_fact_kargs):
        """
        The datatype extractor can be used to determine the data type of
        values based on contextual information. For the generic XML import, we
        know nothing about data types and simply return False in order to
        signify that no data type information could be established.

        """
        return False


    def xml_import(self,
                   filepath=None,
                   xml_content = None,
                   markings=None,
                   identifier_ns_uri = None,
                   uid = None,
                   **kargs):

        """
        Call this function for generic import of an XML file with the following arguments:

        - filename: Path to XML file
        - markings (optional): List of information objects with which the imported object
          should be marked
        - identifier_ns_uri: Namespace of the identifiers, if identifiers are to be created
        - uid (optional): unique identifier -- if none is given, the SHA256 of the file
          contents are used as identifier.
         """


        # Reset bookkeeping dictionaries


        if not markings:
            markings = []
        if not identifier_ns_uri:
            identifier_ns_uri = DINGOS_DEFAULT_ID_NAMESPACE_URI

        self.__init__()

        # Carry out generic XML import
        import_result = DingoImporter.xml_import(xml_fname=filepath,
                                                 xml_content=xml_content,
                                                 ns_mapping=self.namespace_dict,
                                                 embedded_predicate=self.cybox_embedding_pred,
                                                 id_and_revision_extractor=self.id_and_revision_extractor)


        # Extract data required for creating info object

        id_and_rev_info = import_result['id_and_rev_info']
        elt_name = import_result['elt_name']
        elt_dict = import_result['dict_repr']
        file_content = import_result['file_content']

        if uid:
            id_and_rev_info['id'] = uid
        else:
            id_and_rev_info['id'] = hashlib.sha256(file_content).hexdigest()

        id_and_rev_info['timestamp'] = timezone.now()

        create_timestamp = id_and_rev_info['timestamp']

        iobject_family_name = self.namespace_dict.get(elt_dict.get('@@ns', None), DINGOS_GENERIC_FAMILY_NAME)

        iobject_family_revision_name = ''
        iobject_type_name = elt_name
        iobject_type_namespace_uri = self.namespace_dict.get(elt_dict.get('@@ns', None), DINGOS_GENERIC_FAMILY_NAME)

        # Create info object

        DingoImporter.create_iobject(iobject_family_name=iobject_family_name,
                                     iobject_family_revision_name=iobject_family_revision_name,
                                     iobject_type_name=iobject_type_name,
                                     iobject_type_namespace_uri=iobject_type_namespace_uri,
                                     iobject_type_revision_name='',
                                     iobject_data=elt_dict,
                                     uid=id_and_rev_info['id'],
                                     identifier_ns_uri=identifier_ns_uri,
                                     timestamp=id_and_rev_info['timestamp'],
                                     create_timestamp=create_timestamp,
                                     markings=markings,
                                     config_hooks={'special_ft_handler': self.ft_handler_list(),
                                                   'datatype_extractor': self.datatype_extractor},
                                     namespace_dict=self.namespace_dict,
        )


class DingoImportCommand(BaseCommand):
    """
    This class serves as basis for import commands that are specified
    in the management/commands directory of a Django app using DINGO
    for imports.

    It basically adds the following command line arguments and associated
    processing to the Dingo.BaseCommand class:

    - `--marking_json` is used to specify a json file that contains
      data for an information object with which all information objects
      generated by the XML import are to be marked. Here is an example
      the contents of such a json file::

           {"Mechanism" : {"Category":"Commandline Import",
                           "User": "DINGO[_username]",
                           "Commandline": {"Command":"DINGO[_command]",
				           "KeywordArguments":"DINGO[_kargs]",
				           "Arguments":"DINGO[_args]"}
	                      },
          "Source" : "DINGO[source]"}

      As becomes apparent above, the JSON may contain placeholders.
      of form 'DINGO[<placeholder_name>]'. Placeholders with names
      that start with a '_' are filled in automatically -- the definition
      for user-defined placeholders is provided with the
      command-line argument --marking-pfill.

    - `--marking-pfill` takes a list of arguments that are processed pairwise:
      the first component of each pair is interpreted as placeholder name,
      the second as value to fill in for that placeholder.

      To fill in the 'source' placeholder in the above example, you
      might call the command line with::

                  --marking-pfill  source "Regular import from Malware Sandbox"

      Be sure to encompass placeholder values in quotation marks
      (If you are using PyCharm as IDE: note that PyCharm messes up quoted
      commandline arguments in its run configuration -- you have to test
      that stuff from a true commandline).
    - `--id-namespace-uri` stores URI for namespace to be used for qualifying
         the identifiers of the information objects.
    """
    args = 'xml-file xml-file ... (you can use wildcards)'
    help = 'Imports xml files of specified paths into DINGO with generic import'

    RE_SEARCH_PLACEHOLDERS = re.compile(r"DINGO\[(?P<bla>[^\]]+)\]")

    option_list = BaseCommand.option_list + (
        make_option('-m', '--marking_json',
                    action='store',
                    dest='marking_json',
                    default=None,
                    help='File with json representation of information of marking to be associated with imports.'),
        make_option('-p', '--marking_pfill',
                    action='append',
                    nargs=2,
                    default=[],
                    dest='placeholder_fillers',
                    help='Key-value pairs used to fill in placeholders in marking as described in marking file.'),
        make_option('-n','--id_namespace_uri',
                    action='store',
                    default=None,
                    dest='identifier_ns_uri',
                    help='URI of namespace used to qualify the identifiers of the created information objects.'),
    )


    Importer = Generic_XML_Import()


    def __init__(self, *args, **kwargs):
        self.xml_import_function = kwargs.get('import_function', None)
        try:
            del (kwargs['import_function'])
        except:
            pass
        super(DingoImportCommand,self).__init__(*args,**kwargs)

    def create_import_marking(self, args, options):
        """
        This function parses the `--marking_json` and `--marking_pfill` arguments
        and creates a marking object if these parameters are specified.

        Look into 'dingos/management/commands/dingos_generic_xml_import.py to see how
        this is used to specify a Django command line argument 'dingos_generic_xml_import'
        that can be called with Django's 'manage.py'
        """
        marking_json = None
        placeholder_dict = {}

        if options.get('marking_json'):
            # Open json
            with open(options['marking_json'], 'r') as content_file:
                marking_json = content_file.read()

            del(options['marking_json'])

            # Find all placeholders
            placeholders = self.RE_SEARCH_PLACEHOLDERS.findall(marking_json)

            # Create prefilled dictionary so that later when we use
            # the json text as format string we do not encounter problems
            # for undefined placeholders.

            if placeholders:

                for placeholder in placeholders:
                    placeholder_dict[placeholder] = 'EMPTY'

        # Read in command-line specified placeholder values.
        if options.get('placeholder_fillers'):
            for (placeholder, value) in options['placeholder_fillers']:
                placeholder_dict[placeholder] = value

            del(options['placeholder_fillers'])

        # Add standard values (this list may grow in future)
        placeholder_dict['_username'] = getpass.getuser()
        placeholder_dict['_command'] = os.path.basename(__file__)
        placeholder_dict['_kargs'] = "%s" % options
        placeholder_dict['_args'] = args

        if marking_json:
            # Massage the json text such that we can use it as format string
            # to fill in the placeholders

            # Escape possible '%'
            marking_json = marking_json.replace('%','\%')
            # Replace placeholder definitions with python string formatting
            marking_json = self.RE_SEARCH_PLACEHOLDERS.sub("%(\\1)s", marking_json)

            # Use string formatting to fill in placeholders

            marking_json = marking_json % placeholder_dict

            # Finally, parse json
            marking_dict = dict2DingoObjDict(json.loads(marking_json))

            # Create info object for marking
            marking = DingoImporter.create_marking_iobject(metadata_dict=marking_dict)

            return marking

        return None

    def handle(self, *args, **options):
        # The function create_import_marking inherited from
        # DingoImport command is able to create a dictionary
        # structure for a marking with object resulting
        # from the import command will be marked.

        marking = self.create_import_marking(args,options)

        if marking:
            markings = [marking]
        else:
            markings = []

        #if len(args) > 1 and options['identifier']:
        #    raise CommandError('Option --identifier not supported for more than one file per import.')

        if len(args) == 0:
            logger.warning("No files for import specified!")
        else:
            for arg in args:
                logger.info("Starting processing")

                if len(glob.glob(arg)) == 0:
                    logger.warning("No file(s) %s for import found!" % arg)

                for filename in glob.glob(arg):

                    logger.info("Starting import of %s" % filename)

                    self.Importer.xml_import(filepath = filename,
                                             markings = markings,
                                             **options)
